# -*- coding: utf-8 -*-
from functools import wraps as _wraps
from functools import partial
import weakref
from hashlib import md5
from base64 import b64encode
import time
import logging
import random
import pydoc
from .clients import base, asyncache, tiered

try:
    from .mq import coherence
    from .clients import coherent
    no_coherence = False
except ImportError:
    no_coherence = True

from .clients.base import CacheMissError

try:
    import cython
except ImportError:
    # Make cython annotations work without cython
    _cy_noop = lambda f: f
    class _cython:
        locals = lambda **kw: _cy_noop
    globals()['cython'] = _cython

class NO_NAMESPACE:
    pass

class _NONE:
    pass

try:
    import multiprocessing
except ImportError:
    class multiprocessing:
        @staticmethod
        def cpu_count():
            return 1

def wraps(wrapped):
    w = _wraps(wrapped)
    def decor(wrapper):
        wrapper = w(wrapper)
        wrapper.__doc__ = "\n".join(
            pydoc.render_doc(wrapped, 'cached %s:').split('\n', 4)[:3]
            + list(filter(bool, [wrapper.__doc__])))
        return wrapper
    return decor

def _make_namespace(f, salt = None, salt2 = None):
    f = getattr(f, 'im_func', f)
    fname = getattr(f, '__name__', None)
    if fname is None:
        fname = getattr(f, 'func_name', None)
    if fname is None:
        # FTW
        return repr(f)

    mname = getattr(f, '__module__', '')

    fcode = getattr(f, '__code__', None)
    if fcode is None:
        fcode = getattr(f, 'func_code', None)
    if fcode is not None:
        fpath = '%s:%d' % (fcode.co_filename, fcode.co_firstlineno)
    else:
        fpath = ''

    try:
        body_digest = md5(fpath.encode("utf8"))
        if salt:
            body_digest.update(salt.encode("utf8"))
        if salt2:
            body_digest.update(salt2.encode("utf8"))
        if fcode:
            body_digest.update(getattr(fcode, 'co_code', b''))
        return "%s.%s#%s" % (mname,fname,b64encode(body_digest.digest()).rstrip(b"=\n"))
    except Exception as e:
        return repr(f)

def _simple_put_deferred(future, client, f, key, ttl, *p, **kw):
    defer = asyncache.Defer(f, *p, **kw)
    if future is not None:
        defer.future = future
    return client.put(key, defer, ttl)

def _coherent_put_deferred(shared, async_ttl, future, client, f, key, ttl, *p, **kw):
    return client.put_coherently(key, ttl,
        lambda : not shared.contains(key, async_ttl),
        future,
        f, *p, **kw)

class CacheStats(object):
    __slots__ = (
        'hits', 'misses', 'errors',
        'sync_misses', 'sync_errors',
        'min_miss_time', 'max_miss_time', 'sum_miss_time', 'sum_miss_time_sq',
        'miss_time_histogram', 'miss_time_histogram_bins', 'miss_time_histogram_max', 'wait_time',
    )

    def __init__(self):
        self.clear()
        self.set_histogram_bins(0, None)

    def clear(self):
        self.hits = self.misses = self.errors = self.sync_misses = self.sync_errors = 0
        self.max_miss_time = self.sum_miss_time = self.sum_miss_time_sq = self.wait_time = 0
        self.min_miss_time = -1

    def set_histogram_bins(self, bins, maxtime):
        if bins <= 0:
            self.miss_time_histogram_bins = self.miss_time_histogram_max = 0
            self.miss_time_histogram = None
        else:
            self.miss_time_histogram = [0] * bins
            self.miss_time_histogram_bins = bins
            self.miss_time_histogram_max = maxtime

    def add_histo(self, time):
        if self.miss_time_histogram:
            hmax = self.miss_time_histogram_max
            hbins = self.miss_time_histogram_bins
            hbin = min(hbins-1, min(hmax, time) * hbins / hmax)
            self.miss_time_histogram[hbin] += 1

decorated_functions = weakref.WeakSet()

def cached(client, ttl,
        key = lambda *p, **kw:(p,frozenset(kw.items()) or ()),
        namespace = None,
        value_serialization_function = None,
        value_deserialization_function = None,
        async_writer_queue_size = None,
        async_writer_workers = None,
        async_writer_threadpool = None,
        async_writer_kwargs = None,
        async_ttl = None,
        async_client = None,
        async_expire = None,
        lazy_kwargs = {},
        async_lazy_recheck = False,
        async_lazy_recheck_kwargs = {},
        async_processor = None,
        async_processor_workers = None,
        async_processor_threadpool = None,
        async_processor_kwargs = None,
        renew_time = None,
        future_sync_check = None,
        initialize = None,
        decorate = None,
        timings = True,
        ttl_spread = True,
        autonamespace_version_salt = None,
        _eff_async_ttl = None,
        _put_deferred = None,
        _fput_deferred = None,
        _lazy_recheck_put_deferred = None,
        _flazy_recheck_put_deferred = None ):
    """
    This decorator provides cacheability to suitable functions.

    To be considered suitable, the values received on parameters must be immutable types (otherwise caching will
    be unreliable).

    Caches are thread-safe only if the provided clients are thread-safe, no additional safety is provided. If you
    have a thread-unsafe client you want to make safe, use a (ReadWrite)SyncAdapter. Since synchronization adapters
    only apply to store manipuation functions, and not the wrapped function, deadlocks cannot occur.

    The decorated function will provide additional behavior through

    Attributes
    ----------

        client: the backing cache client. The provided client is never used as-is, and instead is wrapped in a
            NamespaceWrapper. This is it.

        ttl: The configured TTL

        async_ttl: The configured async TTL

        clear(): forget all cached values. Since the client might be shared, it will only increase an internal
            revision mark used to decorate keys, so the cache will not be immediately purged. For that, use
            client.clear() (but beware that it will also clear other caches sharing the same client).

        invalidate(...): mimicking the underlying function's signature, it will, instead of invoking the function,
            invalidate cached entries sharing the call's key.

        expire(...): like invalidate, but the key is just marked as out-of-date, requiring immediate refresh,
            but not completely invalid.

        put(_cache_put, ...): mimicking the underlying function's signature after the first positional argument,
            except for one keyword argument _cache_put, it will forcibly set the cached value for that key to
            be what's supplied in _cache_put. Although there will be no invocation of the target function,
            the write will be synchronous unless the underlying cache client is async, and for external caches
            this might still mean a significant delay.

        refresh(...): mimicking the underlying function's signature, it will forcefully invoke the function,
            regardless of cache status, and refresh the cache with the returned value.

        uncached(...): this is the underlying function, undecorated.

        peek(...): mimicking the underlying function's signature, it will query the cache without ever invoking
            the underlying function. If the cache doesn't contain the key, a CacheMissError will be raised.

        get_ttl(...): mimicking the underlying function's signature, it will query the cache without ever invoking
            the underlying function, and return both the result and the ttl. Misses return NONE as value and
            a negative ttl, instead of raising a CacheMissError.

        lazy(...): mimicking the underlying function's signature, it will behave just like a cached function call,
            except that if there is no value, instead of waiting for one to be computed, it will just raise
            a CacheMissError. If the access is async, it will start a background computation. Otherwise, it will
            behave just as peek.

        future(...): probably the preferred way of accessing the cache on an asynchronous app, it will return
            a decorated function that will return futures that will receive the value when done.
            For straight calls, if lazy would not raise a CacheMissError, the future will already contain the value,
            resulting in no delays. If the client is tiered and has remote tiers, it's recommendable to add proper
            lazy_kwargs to avoid this synchronous call blocking. Other forms of access perform similarly, always
            returning a future immediately without blocking. Alternatively, setting future_sync_check to False
            will disable this check and always do it through the processor.

        async(): if the underlying client is async, it will return the decorated function (self). Otherwise, it will
            be another decorated function, created on demand, backed by the same client wrapped in an async adapter.
            As such, it can be used to perform asynchronous operations on an otherwise synchronous function.

        on_promote: subdecorator that registers callbacks that will receive promotion events from this function.

        on_value: subdecorator that registers callbacks that will receive freshly computed values from this function.
            The callback will receive the value as its first argument, and all the original arguments for the call
            that initiated the computation following that argument (including keyword arguments).

        stats: cache statistics, containing:
            hits - number of cache hits
            misses - number of cache misses
            errors - number of exceptions caught
            sync_misses - number of synchronous (blocking) misses
            sync_errors - number of synchronous exceptions (propagated to caller)

            min_miss_time - minimum time spent computing a miss
            max_miss_time - maximum time spent computing a miss
            sum_miss_time - total time spent computing a miss (divide by misses and get an average)
            sum_miss_time_sq - sum of squared times spent computing a miss (to compute standard deviation)

            miss_time_histogram - histogram of times spent computing misses, computed only if histogram
                bins and limits are set.
            miss_time_histogram_bins - number of bins configured
            miss_time_histogram_max - maximum histogram time configured

            wait_time - time spent waiting for async updates, that's sync miss time on async queries

            reset(): clear all statistics
            set_histogram_bins(bins, max): configure histogram collection to use "bins" bins spanning
                from 0 to max. If bins is set to 0, max is ignored, and histogram collection is disabled

            All values are approximate, as no synchroniation is attempted while updating.

            sync_misses and sync_errors are caller-visible misses or exceptions. The difference with
            misses and errors respectively are the number of caller-invisible misses or errors.

    Parameters
    ----------

        client: the cache store client to be used

        ttl: the time, in seconds, during which values remain valid.

        callkey: the given key-computing callable

        renew_time: if not None, the time, in seconds, to add to the TTL when an item is scheduled for
            refresh. This renews the current item at a cost, but prevents concurrent readers from attempting
            their own refresh in a rather simple way, short of using a coherence protocol. Roughly
            equivalent to the dogpile pattern with a timeout as specified.

        key: (optional) A key derivation function, that will take the same arguments as the underlying function,
            and should return a key suitable to the client. If not provided, a default implementation that will
            usually work for most primitive argument types will be used.

        namespace: (optional) If provided, the namespace used to identify cache entries will be the one provided.
            If not, a default one will be derived out of the function's module and name, which may differ between
            platforms, so you'll want to provide a stable one for shared caches. If NO_NAMEPSACE is given,
            no namespace decoration will be applied. Specify if somehow collisions are certain not to occur.

        autonamespace_version_salt: (optional) If provided, it will alter the automatically generated namespace
            in a predictable and stable way. Can be used to force version upgrades when the automatic namespace
            is not able to pick up code changes.

        value_serialization_function: (optional) If provided, values will not be stored directly into the cache,
            but the result of applying this function to them. Use if the cache is remote and does not natively
            support the results given by the underlying function, or if stripping of runtime-specific data is
            required.

        value_deserialization_function: (optional) Counterpart to value_serialization_function, it will be applied
            on values coming from the cache, before returning them as cached function calls.

        async_writer_queue_size: (optional) Writer queue size used for the async() client. Default is 100.

        async_writer_workers: (optional) Number of async workers for the async() client.
            Default is multiprocessing.cpu_count

        async_writer_threadpool: (optional) Threadpool to be used for the async writer, instead of
            workers, this can specify a specific thread pool to use (perhaps shared among other writers).
            It can also be a callable, in which case it must be a factory function that takes the number
            of workers as argument and returns a threadpool to be used. It's recommendable to always use
            factories instead of instances, to avoid premature thread initialization.

        async_writer_kwargs: (optional) Optional arguments to be used when constructing AsyncWriteCacheClient
            wrappers. Ignored if an explicit async_client is given.

        async_client: (optional) Alternatively to async_writer_queue_size and async_writer_workers, a specific
            async client may be specified. It is expected this client will be an async wrapper of the 'client'
            mandatorily specified, that can be shared among decorated functions (to avoid multiplying writer
            threads).

        async_ttl: (optional) The TTL at which an async refresh starts. For example, async_ttl=1 means to start
            a refresh just 1 second before the entry expires. It must be smaller than ttl. Default is half the TTL.
            If negative, it means ttl - async_ttl, which reverses the logic to mean "async_ttl" seconds after
            creation of the cache entry.

        async_processor_workers: (optional) Number of threads used for async cache operations (only applies to
            future() calls, other async operations are configured with async_writer args). Only matters if the
            clients perform expensive serialization (there's no computation involved otherwise). Default is
            multiprocessing.cpu_count

        async_processor_threadpool: (optional) Threadpool to be used for the async processor, instead of
            workers, this can specify a specific thread pool to use (perhaps shared among other processors).
            It can also be a callable, in which case it must be a factory function that takes the number
            of workers as argument and returns a threadpool to be used. It's recommendable to always use
            factories instead of instances, to avoid premature thread initialization.

        async_processor_kwargs: (optional) Optional arguments to be used when constructing AsyncCacheProcessors.
            Ignored if an explicit async_processor is given.

        async_processor: (optional) Shared processor to utilize in future() calls.

        future_sync_check: (optional) Whether to perform a quick synchronous check of the cache with lazy_kwargs
            in order to avoid stalling on the processor for cached access. Set to False if there's no non-blocking
            tier in the given cache client.

        async_expire: (optional) A callable that will get an expired key, when TTL falls below async_ttl.
            Common use case is to pass a first-tier's expire bound method, thus initiating a refresh.

        lazy_kwargs: (optional) kwargs to send to the client's getTtl when doing lazy requests. Useful for
            tiered clients, that can accept access modifiers through kwargs

        async_lazy_recheck: (optional) when sending lazy_kwargs that may result in spurious CacheMissError,
            specifying this on True will trigger an async re-check of the cache (to verify the need for an
            async refresh).

        async_lazy_recheck_kwargs: (optional) when setting async_lazy_recheck, recheck kwargs can be specified
            here (default empty).

        initialize: (optional) A callable hook to be called right before all accesses. It should initialize whatever
            is needed initialized (like daemon threads), and only once (it should be a no-op after it's called once).
            It can return True to avoid being called again (any return value that doesn't evaluate to True will
            be ignored).

        decorate: (optional) A decorator to apply to all call-like decorated functions. Since @cached creates many
            variants of the function, this is a convenience over manually decorating all variants.

        timings: (optional) Whether to gather timing statistics. If true, misses will be timed, and timing data
            will be included in the stats attribute. It does imply some overhead. Default is True.

        ttl_spread: (optional - default True). If None, the TTL will be used as-is for cache insertions.
            If True, an automatic TTL spread will be computed so insertions and recomputatins are better distributed
            in time. If a number (float or int), an equal-type random amount between [-ttl_spread, ttl_spread]
            will be added.
    """
    if value_serialization_function or value_deserialization_function:
        client = base.DecoratedWrapper(client,
            value_decorator = value_serialization_function,
            value_undecorator = value_deserialization_function )
    if namespace is not None and namespace is not NO_NAMESPACE:
        client = base.NamespaceWrapper(namespace, client)
        if async_client:
            async_client = base.NamespaceMirrorWrapper(client, async_client)

    if not client.is_async:
        if async_writer_queue_size is None:
            async_writer_queue_size = 100
        if async_writer_workers is None:
            async_writer_workers = multiprocessing.cpu_count()

    if async_ttl is None:
        async_ttl = ttl / 2
    elif async_ttl < 0:
        async_ttl = ttl + async_ttl

    if ttl_spread is True:
        ttl_spread = min(ttl, async_ttl, abs(ttl-async_ttl)) / 2
    if ttl_spread:
        spread_type = type(ttl_spread)
        eff_ttl = lambda r = random.random : ttl + spread_type(ttl_spread * (r() - 0.5))
    else:
        eff_ttl = lambda : ttl

    if _put_deferred is None:
        _fput_deferred = _simple_put_deferred
        _put_deferred = partial(_fput_deferred, None)
    if _lazy_recheck_put_deferred is None:
        _flazy_recheck_put_deferred = _simple_put_deferred
        _lazy_recheck_put_deferred = partial(_flazy_recheck_put_deferred, None)

    if async_processor is not None and async_processor_workers is None:
        async_processor_workers = multiprocessing.cpu_count()

    if async_writer_kwargs is None:
        async_writer_kwargs = {}
    async_writer_kwargs.setdefault('threadpool', async_writer_threadpool)

    if async_processor_kwargs is None:
        async_processor_kwargs = {}
    async_processor_kwargs.setdefault('threadpool', async_processor_threadpool)

    # Copy to avoid modifying references to caller objects
    elazy_kwargs = lazy_kwargs.copy()
    easync_lazy_recheck_kwargs = async_lazy_recheck_kwargs.copy()
    eget_async_lazy_recheck_kwargs = easync_lazy_recheck_kwargs.copy()

    EMPTY_KWARGS = {}
    _NONE_ = _NONE
    Future = asyncache.Future

    def decor(f):
        if namespace is None:
            salt2 = repr((ttl,))
            nclient = base.NamespaceWrapper(_make_namespace(
                f, salt = autonamespace_version_salt, salt2 = salt2), client)
            if async_client:
                nasync_client = base.NamespaceMirrorWrapper(nclient, async_client)
            else:
                nasync_client = async_client
        else:
            nclient = client
            nasync_client = async_client

        stats = CacheStats()

        if initialize is not None:
            def _initialize():
                nonlocal _initialize
                stop_initializing = initialize()
                if stop_initializing:
                    _initialize = None
        else:
            _initialize = None

        # static async ttl spread, to avoid inter-process contention
        if ttl_spread:
            if _eff_async_ttl:
                eff_async_ttl = _eff_async_ttl
            else:
                eff_async_ttl = max(async_ttl / 2, async_ttl - spread_type(ttl_spread * 0.25 * random.random()))
        else:
            eff_async_ttl = async_ttl

        # Wrap and track misses and timings
        if timings:
            of = f
            @wraps(of)
            @cython.locals(t0=cython.double, t1=cython.double, t=cython.double)
            def af(*p, **kw):
                stats.misses += 1
                try:
                    t0 = time.time()
                    rv = of(*p, **kw)
                    t1 = time.time()
                    t = t1-t0
                    try:
                        if t > stats.max_miss_time:
                            stats.max_miss_time = t
                        if stats.min_miss_time < 0 or stats.min_miss_time > t:
                            stats.min_miss_time = t
                        stats.sum_miss_time += t
                        stats.sum_miss_time_sq += t*t
                        if stats.miss_time_histogram:
                            stats.add_histo(t)
                    except:
                        # Ignore stats collection exceptions.
                        # Quite possible since there is no thread synchronization.
                        pass
                    if value_callbacks:
                        try:
                            _value_callback(rv, *p, **kw)
                        except:
                            # Just log callback exceptions, orthogonal behavior shouldn't propagate to the caller
                            logging.getLogger('chorde').error("Error on value callback", exc_info = True)
                            pass
                    return rv
                except:
                    stats.errors += 1
                    raise
            @wraps(of)
            def f(*p, **kw):
                stats.sync_misses += 1
                return af(*p, **kw)
        else:
            of = f
            @wraps(of)
            def af(*p, **kw):  # lint:ok
                stats.misses += 1
                try:
                    rv = of(*p, **kw)
                    if value_callbacks:
                        try:
                            _value_callback(rv, *p, **kw)
                        except:
                            # Just log callback exceptions, orthogonal behavior shouldn't propagate to the caller
                            logging.getLogger('chorde').getLogger('chorde').error("Error on value callback", exc_info = True)
                            pass
                    return rv
                except:
                    stats.errors += 1
                    raise
            @wraps(of)
            def f(*p, **kw):  # lint:ok
                stats.sync_misses += 1
                rv = af(*p, **kw)
                if value_callbacks:
                    try:
                        _value_callback(rv, *p, **kw)
                    except:
                        # Just log callback exceptions, orthogonal behavior shouldn't propagate to the caller
                        logging.getLogger('chorde').error("Error on value callback", exc_info = True)
                        pass
                return rv

        @wraps(of)
        def cached_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return f(*p, **kw)

            try:
                rv = nclient.get(callkey, **get_kwargs)
                stats.hits += 1
            except CacheMissError:
                rv = f(*p, **kw)
                nclient.put(callkey, rv, eff_ttl())
            return rv
        if decorate is not None:
            cached_f = decorate(cached_f)

        @wraps(of)
        def get_ttl_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return f(*p, **kw)
            rv = nclient.getTtl(callkey, **get_kwargs)
            if rv[1] < 0:
                stats.misses += 1
            else:
                stats.hits += 1
            return rv
        if decorate is not None:
            get_ttl_f = decorate(get_ttl_f)

        @wraps(of)
        @cython.locals(t0=cython.double, t1=cython.double, t=cython.double)
        def async_cached_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return f(*p, **kw)

            client = aclient
            __NONE = _NONE_
            rv, rvttl = client.getTtl(callkey, __NONE, **get_kwargs)

            if (rv is __NONE or rvttl < eff_async_ttl) and not client.contains(callkey, eff_async_ttl):
                if renew_time is not None:
                    if rv is not __NONE:
                        nclient.renew(callkey, eff_async_ttl + renew_time)
                    elif placeholder_value_fn_cell:
                        placeholder = placeholder_value_fn_cell[0](*p, **kw)
                        nclient.add(callkey, placeholder, eff_async_ttl + renew_time)
                        rv = placeholder
                        rvttl = eff_async_ttl + renew_time
                # Launch background update
                _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
            elif rv is not __NONE:
                if rvttl < eff_async_ttl:
                    client.promote(callkey, ttl_skip = eff_async_ttl, **get_kwargs)
                stats.hits += 1

            if rv is __NONE:
                # Must wait for it
                if timings:
                    t0 = time.time()
                client.wait(callkey)
                rv, rvttl = client.getTtl(callkey, __NONE, **get_kwargs)
                if rv is __NONE or rvttl < eff_async_ttl:
                    # FUUUUU
                    rv = f(*p, **kw)
                stats.sync_misses += 1
                if timings:
                    t1 = time.time()
                    t = t1-t0
                    stats.wait_time += t

            return rv
        if decorate is not None:
            async_cached_f = decorate(async_cached_f)

        @wraps(of)
        def future_cached_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return fclient.do_async(f, *p, **kw)

            client = aclient
            clientf = fclient
            frv = Future()
            __NONE = _NONE_

            if future_sync_check:
                # Quick sync call with lazy_kwargs
                rv, rvttl = client.getTtl(callkey, __NONE, **elazy_kwargs)
            else:
                rv = __NONE
                rvttl = -1

            if (rv is __NONE or rvttl < eff_async_ttl):
                # The hard way
                if rv is __NONE:
                    # It was a miss, so wait for setting the value
                    def on_value(value):
                        stats.hits += 1
                        frv.set(value[0])
                        # If it's stale, though, start an async refresh
                        if value[1] < eff_async_ttl and not nclient.contains(callkey, eff_async_ttl,
                                **easync_lazy_recheck_kwargs):
                            if renew_time is not None and (rv is not __NONE or lazy_kwargs):
                                nclient.renew(callkey, eff_async_ttl + renew_time)
                            _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                    def on_miss():
                        if renew_time is not None and placeholder_value_fn_cell:
                            placeholder = placeholder_value_fn_cell[0](*p, **kw)
                            nclient.add(callkey, placeholder, eff_async_ttl + renew_time)
                        _fput_deferred(frv, client, af, callkey, eff_ttl(), *p, **kw)
                    def on_exc(exc_info):
                        stats.errors += 1
                        return frv.exc(exc_info)
                    clientf.getTtl(callkey, ttl_skip = eff_async_ttl, **eget_async_lazy_recheck_kwargs)\
                        .on_any(on_value, on_miss, on_exc)
                else:
                    # It was a stale hit, so set the value now, but start a touch-refresh
                    stats.hits += 1
                    frv._set_nothreads(rv)
                    def on_value(contains):  # lint:ok
                        if not contains:
                            if renew_time is not None and (rv is not __NONE or lazy_kwargs):
                                nclient.renew(callkey, eff_async_ttl + renew_time)
                            _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                        else:
                            # just promote
                            nclient.promote(callkey, ttl_skip = eff_async_ttl, **get_kwargs)
                    def on_miss():  # lint:ok
                        if renew_time is not None and placeholder_value_fn_cell:
                            placeholder = placeholder_value_fn_cell[0](*p, **kw)
                            nclient.add(callkey, placeholder, eff_async_ttl + renew_time)
                        _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                    def on_exc(exc_info):  # lint:ok
                        stats.errors += 1
                    clientf.contains(callkey, eff_async_ttl, **easync_lazy_recheck_kwargs)\
                        .on_any_once(on_value, on_miss, on_exc)
            else:
                stats.hits += 1
                frv._set_nothreads(rv)
            return frv
        if decorate is not None:
            future_cached_f = decorate(future_cached_f)

        @wraps(of)
        def lazy_cached_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                raise CacheMissError

            try:
                rv = nclient.get(callkey, **elazy_kwargs)
                stats.hits += 1
                return rv
            except CacheMissError:
                stats.misses += 1
                raise
        if decorate is not None:
            lazy_cached_f = decorate(lazy_cached_f)
        peek_cached_f = lazy_cached_f

        @wraps(of)
        def future_lazy_cached_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                raise CacheMissError

            client = aclient
            frv = Future()
            __NONE = _NONE_

            if future_sync_check:
                # Quick sync call with lazy_kwargs
                rv, rvttl = client.getTtl(callkey, __NONE, **elazy_kwargs)
            else:
                rv = __NONE
                rvttl = -1

            if (rv is __NONE or rvttl < eff_async_ttl):
                # The hard way
                clientf = fclient
                if rv is __NONE and (not future_sync_check or async_lazy_recheck):
                    # It was a preliminar miss, so wait for a recheck to set the value
                    def on_value(value):
                        if value[1] >= 0:
                            stats.hits += 1
                            frv.set(value[0])
                        else:
                            # Too stale
                            frv.miss()
                        # If it's stale, though, start an async refresh
                        if value[1] < eff_async_ttl and not client.contains(callkey, eff_async_ttl,
                                **easync_lazy_recheck_kwargs):
                            if renew_time is not None and (rv is not __NONE or lazy_kwargs):
                                nclient.renew(callkey, eff_async_ttl + renew_time)
                            _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                    def on_miss():
                        # Ok, real miss, report it and start the computation
                        frv.miss()
                        if renew_time is not None and placeholder_value_fn_cell:
                            placeholder = placeholder_value_fn_cell[0](*p, **kw)
                            nclient.add(callkey, placeholder, eff_async_ttl + renew_time)
                        _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                    def on_exc(exc_info):
                        stats.errors += 1
                        frv.exc(exc_info)
                    clientf.getTtl(callkey, ttl_skip = eff_async_ttl, **eget_async_lazy_recheck_kwargs)\
                        .on_any(on_value, on_miss, on_exc)
                else:
                    # It was a stale hit or permanent miss, so set the value now, but start a touch-refresh
                    if rv is __NONE:
                        stats.misses += 1
                        frv._miss_nothreads()
                    else:
                        stats.hits += 1
                        frv._set_nothreads(rv)
                    def on_value(contains):  # lint:ok
                        if not contains:
                            if renew_time is not None and (rv is not __NONE or lazy_kwargs):
                                nclient.renew(callkey, eff_async_ttl + renew_time)
                            _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                        else:
                            # just promote
                            nclient.promote(callkey, ttl_skip = eff_async_ttl, **get_kwargs)
                    def on_miss():  # lint:ok
                        if renew_time is not None and placeholder_value_fn_cell:
                            placeholder = placeholder_value_fn_cell[0](*p, **kw)
                            nclient.add(callkey, placeholder, eff_async_ttl + renew_time)
                        _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                    def on_exc(exc_info):  # lint:ok
                        stats.errors += 1
                    clientf.contains(callkey, eff_async_ttl, **easync_lazy_recheck_kwargs)\
                        .on_any_once(on_value, on_miss, on_exc)
            else:
                stats.hits += 1
                frv._set_nothreads(rv)
            return frv
        if decorate is not None:
            future_lazy_cached_f = decorate(future_lazy_cached_f)

        @wraps(of)
        def future_peek_cached_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                raise CacheMissError

            client = aclient
            clientf = fclient
            frv = Future()
            __NONE = _NONE_

            if future_sync_check:
                # Quick sync call with lazy_kwargs
                rv, rvttl = client.getTtl(callkey, __NONE, **elazy_kwargs)
            else:
                rv = __NONE
                rvttl = -1

            if (rv is __NONE or rvttl < eff_async_ttl):
                # The hard way
                if rv is __NONE and (not future_sync_check or async_lazy_recheck):
                    # It was a miss, so wait for setting the value
                    def on_value(value):
                        if value[1] >= 0:
                            stats.hits += 1
                            return frv.set(value[0])
                        else:
                            # Too stale
                            stats.misses += 1
                            return frv.miss()
                    def on_miss():
                        # Ok, real miss, report it
                        stats.misses += 1
                        return frv.miss()
                    def on_exc(exc_info):
                        stats.errors += 1
                        return frv.exc(exc_info)
                    clientf.getTtl(callkey, ttl_skip = eff_async_ttl, **eget_async_lazy_recheck_kwargs)\
                        .on_any(on_value, on_miss, on_exc)
                else:
                    # It was a stale hit or permanent miss
                    if rv is __NONE or rvttl < 0:
                        stats.misses += 1
                        frv._miss_nothreads()
                    else:
                        stats.hits += 1
                        frv._set_nothreads(rv)
            else:
                stats.hits += 1
                frv._set_nothreads(rv)
            return frv
        if decorate is not None:
            future_peek_cached_f = decorate(future_peek_cached_f)

        @wraps(of)
        def future_get_ttl_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                raise CacheMissError

            # To-Do: intercept hits/misses and update stats?
            #   (involves considerable overhead...)
            return fclient.getTtl(callkey, **get_kwargs)
        if decorate is not None:
            future_get_ttl_f = decorate(future_get_ttl_f)

        @wraps(of)
        def invalidate_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            nclient.delete(callkey)
        if decorate is not None:
            invalidate_f = decorate(invalidate_f)

        @wraps(of)
        def future_invalidate_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            return fclient.delete(callkey)
        if decorate is not None:
            future_invalidate_f = decorate(future_invalidate_f)

        @wraps(of)
        def expire_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            nclient.expire(callkey)
        if decorate is not None:
            expire_f = decorate(expire_f)

        @wraps(of)
        def async_expire_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            aclient.expire(callkey)
        if decorate is not None:
            async_expire_f = decorate(async_expire_f)

        @wraps(of)
        def future_expire_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            return fclient.expire(callkey)
        if decorate is not None:
            future_expire_f = decorate(future_expire_f)

        @wraps(of)
        def put_f(*p, **kw):
            value = kw.pop('_cache_put')
            put_kwargs = kw.pop('_cache_put_kwargs', None)
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            nclient.put(callkey, value, eff_ttl(), **(put_kwargs or EMPTY_KWARGS))
        if decorate is not None:
            put_f = decorate(put_f)

        @wraps(of)
        def async_put_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            value = kw.pop('_cache_put')
            put_kwargs = kw.pop('_cache_put_kwargs', None)
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            aclient.put(callkey, value, eff_ttl(), **(put_kwargs or EMPTY_KWARGS))
        if decorate is not None:
            async_put_f = decorate(async_put_f)

        @wraps(of)
        def future_put_f(*p, **kw):
            value = kw.pop('_cache_put')
            put_kwargs = kw.pop('_cache_put_kwargs', None)
            try:
                callkey = key(*p, **kw)
            except:
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return
            return fclient.put(callkey, value, eff_ttl(), **(put_kwargs or EMPTY_KWARGS))
        if decorate is not None:
            future_put_f = decorate(future_put_f)

        @wraps(of)
        def async_lazy_cached_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                raise CacheMissError

            __NONE = _NONE_
            client = aclient

            rv, rvttl = client.getTtl(callkey, __NONE, **elazy_kwargs)

            if (rv is __NONE or rvttl < eff_async_ttl) and not client.contains(callkey, eff_async_ttl, **elazy_kwargs):
                if async_lazy_recheck:
                    stats.misses += 1

                    # send a Defer that touches the client with recheck kwargs
                    # before doing the refresh. Needs not be coherent.
                    def touch_key(*p, **kw):
                        rv, rvttl = nclient.getTtl(callkey, __NONE, ttl_skip = eff_async_ttl,
                            **eget_async_lazy_recheck_kwargs)
                        if (rv is __NONE or rvttl < eff_async_ttl) and not nclient.contains(callkey, eff_async_ttl,
                                **easync_lazy_recheck_kwargs):
                            if renew_time is not None and (rv is not __NONE or lazy_kwargs):
                                nclient.renew(callkey, eff_async_ttl + renew_time)
                            _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
                        return base.NONE

                    # This will make contains return True for this key, until touch_key returns
                    # This is actually good, since it'll result in immediate misses from now on,
                    # avoiding trying to queue up touch after touch
                    _lazy_recheck_put_deferred(client, touch_key, callkey, ttl, *p, **kw)
                else:
                    _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
            elif rv is not __NONE:
                if rvttl < eff_async_ttl:
                    if async_expire:
                        async_expire(callkey)
                    else:
                        # means client.contains(callkey, eff_async_ttl), so promote
                        client.promote(callkey, ttl_skip = eff_async_ttl, **get_kwargs)
                stats.hits += 1
            else:
                stats.misses += 1

            if rv is __NONE:
                raise CacheMissError(callkey)
            else:
                return rv
        if decorate is not None:
            async_lazy_cached_f = decorate(async_lazy_cached_f)

        @wraps(of)
        def refresh_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return

            rv = f(*p, **kw)
            nclient.put(callkey, rv, eff_ttl())
            return rv
        if decorate is not None:
            refresh_f = decorate(refresh_f)

        @wraps(of)
        def async_refresh_f(*p, **kw):
            if _initialize is not None:
                _initialize()
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return

            client = aclient
            if not client.contains(callkey, 0):
                _put_deferred(client, af, callkey, eff_ttl(), *p, **kw)
        if decorate is not None:
            async_refresh_f = decorate(async_refresh_f)

        @wraps(of)
        def future_refresh_f(*p, **kw):
            try:
                callkey = key(*p, **kw)
            except:
                # Bummer
                logging.getLogger('chorde').error("Error evaluating callkey", exc_info = True)
                stats.errors += 1
                return

            frv = Future()
            _fput_deferred(frv, aclient, af, callkey, eff_ttl(), *p, **kw)
            return frv
        if decorate is not None:
            future_refresh_f = decorate(future_refresh_f)

        if client.is_async:
            cached_f = async_cached_f
            lazy_cached_f = async_lazy_cached_f
        elif nasync_client:
            aclient = nasync_client
        else:
            aclient = None

        promote_callbacks = []
        value_callbacks = []
        get_kwargs = {}
        def _promote_callback(*p, **kw):
            for cb in promote_callbacks:
                cb(*p, **kw)
        def _value_callback(*p, **kw):
            for cb in value_callbacks:
                cb(*p, **kw)
        def on_promote_f(callback):
            promote_callbacks.append(callback)
            get_kwargs.setdefault('promote_callback', _promote_callback)
            elazy_kwargs.setdefault('promote_callback', _promote_callback)
            eget_async_lazy_recheck_kwargs.setdefault('promote_callback', _promote_callback)
            return callback
        def on_value_f(callback):
            value_callbacks.append(callback)
            return callback

        placeholder_value_fn_cell = []
        def placeholder_value_f(fn):
            placeholder_value_fn_cell[:] = [fn]

        fclient = None
        def future_f(initialize = True):
            nonlocal fclient
            if _initialize is not None:
                _initialize()
            if fclient is None and initialize:
                if aclient:
                    _client = aclient
                else:
                    async_f() # initializes aclient
                    _client = aclient

                if async_processor:
                    _client = async_processor.bound(_client)
                else:
                    _client = asyncache.AsyncCacheProcessor(async_processor_workers, _client,
                        **async_processor_kwargs)
                # atomic
                fclient = _client
                future_cached_f.client = fclient
            return future_cached_f

        if not client.is_async:
            def async_f(initialize = True):
                nonlocal aclient
                if _initialize is not None:
                    _initialize()
                if aclient is None and initialize:
                    # atomic
                    aclient = asyncache.AsyncWriteCacheClient(nclient,
                        async_writer_queue_size,
                        async_writer_workers,
                        **async_writer_kwargs)
                    async_cached_f.client = aclient
                return async_cached_f
            async_cached_f.clear = nclient.clear
            async_cached_f.client = aclient if aclient is not None else None
            async_cached_f.bg = weakref.ref(async_cached_f)
            async_cached_f.lazy = async_lazy_cached_f
            async_cached_f.refresh = async_refresh_f
            async_cached_f.peek = peek_cached_f
            async_cached_f.invalidate = invalidate_f
            async_cached_f.expire = async_expire_f
            async_cached_f.uncached = of
            async_cached_f.put = async_put_f
            async_cached_f.ttl = ttl
            async_cached_f.async_ttl = async_ttl
            async_cached_f.callkey = key
            async_cached_f.stats = stats
            async_cached_f.get_ttl = get_ttl_f
            async_cached_f._promote_callbacks = promote_callbacks
            async_cached_f._value_callbacks = value_callbacks
            async_cached_f.on_promote = on_promote_f
            async_cached_f.on_value = on_value_f
            async_cached_f.placeholder_value = placeholder_value_f
            cached_f.bg = async_f
            cached_f.lazy = lazy_cached_f
            cached_f.refresh = refresh_f
            cached_f.peek = peek_cached_f
            cached_f.invalidate = invalidate_f
            cached_f.expire = expire_f
            cached_f.put = put_f
        else:
            aclient = nclient
            cached_f.bg = async_f = weakref.ref(cached_f)
            cached_f.lazy = async_lazy_cached_f
            cached_f.refresh = async_refresh_f
            cached_f.peek = peek_cached_f
            cached_f.invalidate = invalidate_f
            cached_f.expire = expire_f
            cached_f.put = async_put_f

        cached_f.future = future_f
        cached_f.clear = nclient.clear
        cached_f.client = nclient
        cached_f.ttl = ttl
        cached_f.async_ttl = async_ttl or ttl
        cached_f.callkey = key
        cached_f.stats = stats
        cached_f.uncached = of
        cached_f.get_ttl = get_ttl_f
        cached_f._promote_callbacks = promote_callbacks
        cached_f._value_callbacks = value_callbacks
        cached_f.on_promote = on_promote_f
        cached_f.on_value = on_value_f
        cached_f.placeholder_value = placeholder_value_f

        future_cached_f.clear = lambda : fclient.clear()
        future_cached_f.client = None
        future_cached_f.bg = cached_f.bg
        future_cached_f.lazy = future_lazy_cached_f
        future_cached_f.refresh = future_refresh_f
        future_cached_f.peek = future_peek_cached_f
        future_cached_f.invalidate = future_invalidate_f
        future_cached_f.expire = future_expire_f
        future_cached_f.put = future_put_f
        future_cached_f.ttl = ttl
        future_cached_f.async_ttl = async_ttl
        future_cached_f.callkey = key
        future_cached_f.stats = stats
        future_cached_f.uncached = of
        future_cached_f.get_ttl = future_get_ttl_f
        future_cached_f._promote_callbacks = promote_callbacks
        future_cached_f._value_callbacks = value_callbacks
        future_cached_f.on_promote = on_promote_f
        future_cached_f.on_value = on_value_f
        future_cached_f.placeholder_value = placeholder_value_f

        decorated_functions.add(cached_f)
        return cached_f
    return decor

if not no_coherence:

    def coherent_cached(private, shared, ipsub, ttl,
            key = lambda *p, **kw:(p,frozenset(kw.items()) or ()),
            tiered_ = None,
            namespace = None,
            coherence_namespace = None,
            coherence_encoding = 'pyobj',
            coherence_timeout = None,
            value_serialization_function = None,
            value_deserialization_function = None,
            async_writer_queue_size = None,
            async_writer_workers = None,
            async_writer_threadpool = None,
            async_writer_kwargs = None,
            async_ttl = None,
            async_expire = None,
            lazy_kwargs = {},
            async_lazy_recheck = False,
            async_lazy_recheck_kwargs = {},
            async_processor = None,
            async_processor_workers = None,
            async_processor_threadpool = None,
            async_processor_kwargs = None,
            renew_time = None,
            future_sync_check = None,
            initialize = None,
            decorate = None,
            tiered_opts = None,
            ttl_spread = True,
            wait_time = None,
            autonamespace_version_salt = None,
            **coherence_kwargs ):
        """
        This decorator provides cacheability to suitable functions, in a way that maintains coherency across
        multiple compute nodes.

        For suitability considerations and common parameters, refer to cached. The following describes the
        aspects specific to the coherent version.

        The decorated function will provide additional behavior through attributes:
            coherence: the coherence manager created for this purpse

            ipsub: the given IPSub channel

        Params
            ipsub: An IPSub channel that will be used to publish and subscribe to update events.

            private: The private (local) cache client, the one that needs coherence.

            shared: The shared cache client, that reflects changes made by other nodes, or a tuple
                to specify multiple shared tiers.

            tiered_: (optional) A client that queries both, private and shared. By default, a TieredInclusiveClient
                is created with private and shared as first and second tier, which should be the most common case.
                The private client will be wrapped in an async wrapper if not async already, to be able to execute
                the coherence protocol asynchronously. This should be adequate for most cases, but in some,
                it may be beneficial to provide a custom client.

            tiered_opts: (optional) When using the default-constructed tiered client, you can pass additional (keyword)
                arguments here.

            coherence_namespace: (optional) There is usually no need to specify this value, the namespace in use
                by caching will be used for messaging as well, or if caching uses NO_NAMESPACE, the default that
                would be used instead. However, for really high-volume channels, sometimes it is beneficial to pick
                a more compact namespace (an id formatted with struct.pack for example).

            coherence_encoding: (optional) Keys will have to be transmitted accross the channel, and this specifies
                the encoding that will be used. The default 'pyobj' should work most of the time, but it has to
                be initialized, and 'json' or others could be more compact, depending on keys.
                (see CoherenceManager for details on encodings)

            coherence_timeout: (optional) Time (in ms) of peer silence that will be considered abnormal. Default
                is 2000ms, which is sensible given the IPSub protocol. You may want to increase it if node load
                creates longer hiccups.

            wait_time: (optional) Time (in ms) a call that is being computed externally will wait blocking for
                a result. Waiting forever (None, the default) could block the writer threadpool, so it's best to
                specify a reasonably adequate (for the application) timeout.

            Any extra argument are passed verbatim to CoherenceManager's constructor.
        """
        if async_ttl is None:
            async_ttl = ttl / 2
        elif async_ttl < 0:
            async_ttl = ttl + async_ttl

        if ttl_spread is True:
            ttl_spread = min(ttl, async_ttl, abs(ttl-async_ttl)) / 2

        if not private.is_async:
            if async_writer_queue_size is None:
                async_writer_queue_size = 100
            if async_writer_workers is None:
                async_writer_workers = multiprocessing.cpu_count()

        if async_writer_kwargs is None:
            async_writer_kwargs = {}
        async_writer_kwargs.setdefault('threadpool', async_writer_threadpool)

        def decor(f):
            if ttl_spread:
                spread_type = type(ttl_spread)
                eff_async_ttl = max(async_ttl / 2, async_ttl - spread_type(ttl_spread * 0.25 * random.random()))
            else:
                eff_async_ttl = None

            salt2 = repr((ttl,))
            if coherence_namespace is None:
                _coherence_namespace = _make_namespace(f, salt = autonamespace_version_salt, salt2 = salt2)
            else:
                _coherence_namespace = coherence_namespace

            if namespace is None:
                _namespace = _make_namespace(f, salt = autonamespace_version_salt, salt2 = salt2)
            else:
                _namespace = namespace

            if not private.is_async:
                nprivate = asyncache.AsyncWriteCacheClient(private,
                    async_writer_queue_size,
                    async_writer_workers,
                    **async_writer_kwargs)
            else:
                nprivate = private

            if not isinstance(shared, tuple):
                shareds = (shared,)
                sharedt = shared
            elif len(shared) == 1:
                shareds = shared
                sharedt = shared[0]
            else:
                shareds = shared
                sharedt = tiered.TieredInclusiveClient(*shareds, **(tiered_opts or {}))

            if tiered_ is None:
                ntiered = tiered.TieredInclusiveClient(nprivate, *shareds, **(tiered_opts or {}))
            else:
                ntiered = tiered_

            if _namespace is not NO_NAMESPACE:
                ntiered = base.NamespaceWrapper(_namespace, ntiered)
                nprivate = base.NamespaceMirrorWrapper(ntiered, nprivate)
                nshared = base.NamespaceMirrorWrapper(ntiered, sharedt)
            else:
                nshared = sharedt

            coherence_manager = coherence.CoherenceManager(
                _coherence_namespace, nprivate, nshared, ipsub,
                encoding = coherence_encoding,
                **coherence_kwargs)

            nclient = coherent.CoherentWrapperClient(ntiered, coherence_manager, coherence_timeout)

            expired_ttl = async_ttl
            if eff_async_ttl:
                expired_ttl = max(expired_ttl, eff_async_ttl)
            if renew_time:
                expired_ttl = max(expired_ttl, (eff_async_ttl or async_ttl) + renew_time)

            rv = cached(nclient, ttl,
                namespace = NO_NAMESPACE, # Already covered
                key = key,
                value_serialization_function = value_serialization_function,
                value_deserialization_function = value_deserialization_function,
                async_writer_queue_size = async_writer_queue_size,
                async_writer_workers = async_writer_workers,
                async_writer_threadpool = async_writer_threadpool,
                async_writer_kwargs = async_writer_kwargs,
                async_ttl = async_ttl,
                async_expire = async_expire,
                initialize = initialize,
                decorate = decorate,
                lazy_kwargs = lazy_kwargs,
                async_lazy_recheck = async_lazy_recheck,
                async_lazy_recheck_kwargs = async_lazy_recheck_kwargs,
                async_processor = async_processor,
                async_processor_workers = async_processor_workers,
                async_processor_threadpool = async_processor_threadpool,
                async_processor_kwargs = async_processor_kwargs,
                renew_time = renew_time,
                future_sync_check = future_sync_check,
                ttl_spread = ttl_spread,
                _eff_async_ttl = eff_async_ttl,
                _put_deferred = partial(_coherent_put_deferred, nshared, expired_ttl, None,
                    expire_private = nprivate.expire),
                _fput_deferred = partial(_coherent_put_deferred, nshared, expired_ttl,
                    wait_time=wait_time,
                    expire_private = nprivate.expire) )(f)
            rv.coherence = coherence_manager
            rv.ipsub = ipsub
            return rv
        return decor



cdef class Future:
    cdef list _cb
    cdef object _value, _logger, _done_event, __weakref__
    cdef int _running, _cancel_pending, _cancelled

    cdef int c_done(self) except -1
    cdef c_result(self, timeout, int norecurse)

    cpdef _set_nothreads(self, value)
    cpdef set(self, value)
    cpdef _on_stuff(self, callback)



# -*- coding: utf-8 -*-
import collections
import functools
import itertools
import logging
import multiprocessing
import os
import sys
import _thread as thread
import threading
import time
import weakref

from . import worker

class TimeoutError(Exception):
    pass
class TerminateWorker(Exception):
    pass

class WorkerThread(threading.Thread):
    def __init__(self,target,*args,**kwargs):
        threading.Thread.__init__(self, name=kwargs.pop('name', None))
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.__terminate = False

    def run(self, TerminateWorker=TerminateWorker):
        while not self.__terminate:
            try:
                try:
                    self.target(*self.args,**self.kwargs)
                finally:
                    worker_ = worker
                    if worker_ is None:
                        # The None check is necessary to avoid errors during interpreter shutdown
                        self.logger.info("Interpreter shutdown in progress, terminating worker thread")
                        raise TerminateWorker()
                    worker._callCleanupHooks()
            except TerminateWorker:
                self.logger.info("Worker terminated")
                self.terminate(False)
            except Exception:
                self.logger.error("Exception ocurred in worker thread:", exc_info = True)

    def start(self):
        global _nothreads
        threading.Thread.start(self)

    def terminate(self, wait = True):
        self.__terminate = True
        if wait:
            self.join()

try:
    from .clients._async import ExceptionWrapper
except ImportError:
    class ExceptionWrapper(object):  # lint:ok
        __slots__ = ('exc',)

        def __init__(self, exc):
            self.exc = exc

        def reraise(self, strip=True):
            exc = self.value
            if strip:
                del self.value
            try:
                exc_typ, exc_obj, exc_tb = exc
            finally:
                # Don't leave references to the exc/tb in the frame
                del exc
            try:
                if not strip:
                    # Can't raise the same exception object multiple times,
                    # tracebacks accumulate and leak. Use proper chaining.
                    raise exc_typ(*exc_obj.args) from exc_obj
                elif exc_tb is not None:
                    if exc_obj is not None:
                        if getattr(exc_obj, '__traceback__') is not exc_tb:
                            exc_obj = exc_obj.with_traceback(exc_tb)
                        raise exc_obj
                    else:
                        raise exc_typ().with_traceback(exc_tb)
                elif exc_obj is not None:
                    raise exc_obj
                else:
                    raise exc_typ()
            finally:
                # Don't leave references to the exc/tb in the frame
                del exc_typ, exc_obj, exc_tb

class WaitIter:
    def __init__(self, event, timeout = None):
        self.event = event
        self.timeout = timeout
        self._terminate = False
    def terminate(self):
        self._terminate = True
        self.event.set()
    def __iter__(self):
        return self
    def __next__(self, TerminateWorker=TerminateWorker):
        if self._terminate:
            raise TerminateWorker()
        self.event.wait(self.timeout)
        raise StopIteration

class ThreadPool:
    """
    Re-implementation of multiprocessing.pool.ThreadPool optimized for threads
    and asynchronous result-less tasks.
    Implements quasi-lockless double-buffering so queue insertions are very fast, and
    multiple queues for managing fairness.

    This implementation is forcibly a daemon thread pool, which when destroyed
    will cancel all pending tasks, and no task returns any result, and has been
    optimized for that usage pattern.
    """

    Process = WorkerThread

    def __init__(self, workers = None, min_batch = 10, max_batch = 1000, max_slice = None, logger = None,
            name_pattern = None):
        if workers is None:
            workers = multiprocessing.cpu_count()

        self.workers = workers
        self.name_pattern = name_pattern
        self.logger = logger if logger is not None else logging.getLogger('chorde')
        self.__last_worker_index = 1
        self.__workers = None
        self.__pid = os.getpid()
        self.__spawnlock = threading.Lock()
        self.__swap_lock = threading.Lock()
        self.__not_empty = threading.Event()
        self.__empty = threading.Event()
        self.__empty.set()
        self.__cleanup_callbacks = []

        self.local = threading.local()
        self.queues = collections.defaultdict(list)
        self.queue_weights = {}
        self.completed_tasks = 0
        self.__queue_slices = {}
        self.__worklen = 0
        self.__workset = set()
        self.__busyqueues = set()
        self.__busyfactors = {}
        self.__exhausted_iter = WaitIter(self.__not_empty)
        self.__dequeue = self.__exhausted = self.__exhausted_iter.__next__

        self.min_batch = min_batch
        self.max_batch = max_batch
        self.max_slice = max_slice

    def queuelen(self, queue = None):
        return (
            len(self.queues.get(queue,()))
            - self.__queue_slices.get(queue,0)
            + int(self.__worklen * self.__busyfactors.get(queue,0))
        )

    # alias for multiprocessing.pool compatibility
    qsize = queuelen

    # alias for multiprocessing.pool compatibility
    @property
    def _taskqueue(self):
        return self

    def queueprio(self, queue = None):
        return self.queue_weights.get(queue,1)

    def set_queueprio(self, prio, queue = None):
        self.queue_weights[queue] = prio

    def register_cleanup_callback(self, callback):
        self.__cleanup_callbacks.append(callback)

    def _call_cleanup_callbacks(self):
        for callback in self.__cleanup_callbacks:
            try:
                callback()
            except:
                self.logger.error("Error in task cleanup callback", exc_info = True)

    def __swap_queues(self, max=max, min=min, len=len):
        queues = self.queues
        qget = queues.get
        queue_slices = self.__queue_slices
        pget = queue_slices.get
        ppop = queue_slices.pop
        qprio = self.queue_weights.get
        qnames = list(queues.keys())
        wqueues = []
        wprios = []
        iquantities = {}
        itotal = 0

        if qnames:
            # Compute batch size
            # Must be fair, so we must calibrate the batch ends
            # with all queues more or less at the same time
            # Allow some unfairness (but only some)
            # Slices are zero-copy (we don't want to copy a lot of stuff)
            # until maxbatch (we don't want race coditions that result in infinite growth)
            # maxbatch, if None, means "half the queue" (which auto-amortizes the cost of slicing)
            min_batch = self.min_batch
            max_batch = self.max_batch
            max_slice = self.max_slice
            qslots = int(min(max_batch, max(min_batch, min([
                (len(qget(q)) or max_batch) / qprio(q,1)
                for q in qnames
            ]))))
            for qname in qnames:
                q = qget(qname)
                if not q:
                    continue

                qpos = pget(qname,0)
                prio = qprio(qname,1)
                batch = qslots * prio

                # copy-slicing
                qslice = q[qpos:qpos+batch]
                qlen = len(qslice)
                if qlen:
                    iquantities[qname] = qlen
                    itotal += qlen
                    wqueues.append(qslice)
                    wprios.append(prio)
                del qslice

                if qlen < batch or qpos > (max_slice or (len(q)//2)):
                    del q[:qpos+qlen]
                    if qpos:
                        ppop(qname,None)
                else:
                    # move slice position
                    queue_slices[qname] = qpos+qlen

        if wqueues:
            self.__busyqueues.clear()
            self.__busyqueues.update(qnames)

            # Flatten with weights
            # Do it repeatedly to catch stragglers (those that straggle past the flattening step)
            iqueue = []
            iappend = iqueue.append
            islice = itertools.islice
            cycle = itertools.cycle
            repeat = itertools.repeat
            partial = functools.partial

            queues = []
            for q,qprio in zip(wqueues, wprios):
                queues.append(partial(repeat, iter(q).__next__, qprio))

            ioffs = 0
            while queues:
                try:
                    for ioffs,q in islice(cycle(enumerate(queues)), ioffs, None):
                        for q in q():
                            iappend(q())
                except StopIteration:
                    del queues[ioffs]

            self.__worklen = len(iqueue)
            self.__dequeue = iter(iqueue).__next__
            if itotal:
                ftotal = float(itotal)
                self.__busyfactors = {qname: quant/ftotal for qname,quant in iquantities.items()}
            else:
                self.__busyfactors = {}
        elif self.__dequeue is not self.__exhausted:
            self.__not_empty.clear()
            self.__worklen = 0
            self.__busyfactors = {}
            self.__dequeue = self.__exhausted

            # Try again
            # This is a transition from working to empty, which means
            # until now, pushing threads didn't set the weakeup call event.
            # So, before actually sleeping, try again
            self.__swap_queues()
        elif self.__not_empty.isSet():
            # Try again
            # Someone may have sneaked in while we were in the above case
            # and may still neak in after we clear the not_empty event
            self.__not_empty.clear()
            self.__worklen = 0
            self.__busyfactors = {}
            self.__dequeue = self.__exhausted
            self.__swap_queues()
        else:
            # Still empty, can safely give up until signaled
            # Wake up threads trying to join
            self.__empty.set()

    def _dequeue(self, TerminateWorker=TerminateWorker):
        tid = thread.get_ident()
        workset = self.__workset
        termcount = 0
        while True:
            if (self.__dequeue is self.__exhausted and not self.__not_empty.isSet()
                    and (not self.queues or not any(self.queues.values()))):
                # Sounds like there's nothing to do
                # Yeah, gonna wait
                workset.discard(tid)
                if not workset and (not self.queues or not any(self.queues.values())):
                    self.__empty.set()
            else:
                workset.add(tid)
            try:
                rv = self.__dequeue()
                self.__worklen -= 1 # not atomic, but we don't care
                return rv
            except (TerminateWorker, StopIteration) as e:
                # Exhausted whole workqueue?
                with self.__swap_lock:
                    try:
                        if self.__dequeue is self.__exhausted:
                            # Pointless to wait, just swap again
                            if termcount > 0:
                                raise TerminateWorker()
                            else:
                                # First time we get the terminate signal, we try to swap queues
                                # to flush any queued tasks. If we get 2 consecutive signals,
                                # that means we're done with the queue.
                                if isinstance(e, TerminateWorker):
                                    termcount += 1
                                raise StopIteration
                        else:
                            # Try it
                            workset.add(tid)
                            rv = self.__dequeue()
                            self.__worklen -= 1 # not atomic, but we don't care
                            return rv
                    except StopIteration:
                        # Yep, exhausted queue, build up new workqueue
                        self.__swap_queues()
                    except TerminateWorker:
                        # Wake up others so they check, if they're sleeping
                        workset.discard(tid)
                        self.__not_empty.set()
                        raise

    def _enqueue(self, queue, task):
        self.queues[queue].append(task)

        # Wake up waiting threads
        # Note that it's not necessary to invoke this all the time. If the
        # flags are the right way at any point within this function being
        # run, then it already means the respective waiting threads have
        # woken up (or are in the process of waking up) in time to pick up the
        # just-queued value, so avoid the actual operation
        # (which is much more expensive than checking)
        not_empty = self.__not_empty
        if not not_empty.isSet():
            not_empty.set()
        empty = self.__empty
        if empty.isSet():
            empty.clear()

        self.assert_started()

    @staticmethod
    def worker(self, TerminateWorker=TerminateWorker):
        self = self()
        if self is None:
            raise TerminateWorker()

        task = self._dequeue()
        local = self.local
        if task is not None:
            try:
                local.working = True
                task()

                # really couldn't care less about thread safety here ;)
                self.completed_tasks += 1
                if self.completed_tasks > (1<<30):
                    self.completed_tasks = 0
            finally:
                try:
                    self._call_cleanup_callbacks()
                except:
                    self.logger.error("Error in task cleanup hook", exc_info = True)
                try:
                    del local.working
                except:
                    pass

    def in_worker(self):
        return getattr(self.local, 'working', False)

    def is_started(self):
        return not(self.__workers is None or self.__pid != os.getpid())

    def check_started(self):
        return self.is_started() and all([t.is_alive() for t in self.__workers])

    def stop(self, wait = False):
        if self.__workers:
            with self.__spawnlock:
                for w in self.__workers or ():
                    try:
                        if w.is_alive():
                            w.terminate(False)
                    except:
                        pass
                self.__workers = None

            # Wake up threads so they die awake
            self.__not_empty.set()

    def close(self):
        # Signal idle threads to commit suicide
        self.__exhausted_iter.terminate()

    def terminate(self):
        self.stop()

    def start(self):
        if not self.is_started():
            self.populate_workers()

    def assert_started(self):
        if not self.is_started():
            self.populate_workers()

    def join(self, timeout = None):
        if not self.is_started():
            return True
        if timeout is not None:
            now = time.time()
            timeout += now
        while timeout is None or now < timeout:
            if timeout is not None:
                wait_timeout = timeout - now
            elif self.__exhausted_iter._terminate:
                # If the pool is shut down in a way that aborts all queued tasks, we can't
                # merely wait for the empty event to be set, we have to also monitor suiciding threads
                if not self.is_started():
                    break
                wait_timeout = 0.1
            else:
                wait_timeout = None
            if self.__empty.wait(wait_timeout):
                # The event is not 100% certain, we can still get awakened when there's work to do
                # We have to check under __swap_lock to be sure
                with self.__swap_lock:
                    if (self.__dequeue is self.__exhausted and not self.__workset
                            and (not self.queues or not any(self.queues.values()))):
                        return True
                    else:
                        # False alarm, clear it so we don't spin
                        self.__empty.clear()
            else:
                # Timeout
                if timeout is not None or not self.__workset:
                    return (self.__dequeue is self.__exhausted and not self.__workset
                        and (not self.queues or not any(self.queues.values())))
            if timeout is not None:
                now = time.time()
        return not self.is_started() or (self.__dequeue is self.__exhausted and not self.__workset
            and (not self.queues or not any(self.queues.values())))

    def populate_workers(self):
        with self.__spawnlock:
            if not self.is_started():
                name_pattern = self.name_pattern
                index_base = self.__last_worker_index
                self.__last_worker_index += self.workers
                self.__workers = [ self.Process(
                        functools.partial(self.worker, weakref.ref(self)),
                        name = name_pattern % (i+index_base,) if name_pattern is not None else None
                    ) for i in range(self.workers) ]
                for w in self.__workers:
                    w.logger = self.logger
                    w.daemon = True
                    w.start()

                self.__pid = os.getpid()
            # Else, just keep number of workers in sync
            elif len(self.__workers) < self.workers:
                name_pattern = self.name_pattern
                new_workers = self.workers - len(self.__workers)
                index_base = self.__last_worker_index
                self.__last_worker_index += new_workers
                nworkers = [ self.Process(
                        functools.partial(self.worker, weakref.ref(self)),
                        name = name_pattern % (i+index_base,) if name_pattern is not None else None
                    ) for i in range(new_workers) ]
                for w in nworkers:
                    w.daemon = True
                    w.start()
                self.__workers.extend(nworkers)
            elif len(self.__workers) > self.workers:
                nworkers = self.__workers[self.workers:]
                del self.__workers[self.workers:]
                for w in nworkers:
                    w.terminate(False)

    def apply_async(self, task, args = (), kwargs = {}, queue = None):
        if args or kwargs:
            task = functools.partial(task, *args, **kwargs)
        self._enqueue(queue, task)

    def apply(self, task, args = (), kwargs = {}, queue = None, timeout = None):
        rv = None
        ev = threading.Event()
        def stask():
            nonlocal rv
            try:
                rv = task(*args, **kwargs)
            except:
                rv = ExceptionWrapper(sys.exc_info())
            ev.set()
        self._enqueue(queue, stask)
        if ev.wait(timeout):
            if isinstance(rv, ExceptionWrapper):
                rv.reraise()
            else:
                return rv
        else:
            raise TimeoutError

    def subqueue(self, queue, *p, **kw):
        return SubqueueWrapperThreadPool(self, queue, *p, **kw)

class SubqueueWrapperThreadPool:
    """
    Re-implementation of multiprocessing.pool.ThreadPool optimized for threads
    and asynchronous result-less tasks.
    Implements quasi-lockless double-buffering so queue insertions are very fast, and
    multiple queues for managing fairness.

    This implementation is forcibly a daemon thread pool, which when destroyed
    will cancel all pending tasks, and no task returns any result, and has been
    optimized for that usage pattern.
    """

    def __init__(self, pool, queue, priority = None):
        self.queue = queue
        self.pool = pool
        if priority is not None:
            self.set_queueprio(priority)

    def queuelen(self):
        return self.pool.queuelen(self.queue)

    # alias for multiprocessing.pool compatibility
    qsize = queuelen

    # alias for multiprocessing.pool compatibility
    @property
    def _taskqueue(self):
        return self

    @property
    def local(self):
        return self.pool.local

    def in_worker(self):
        return self.pool.in_worker()

    def queueprio(self):
        return self.pool.queueprio(self.queue)

    def set_queueprio(self, prio):
        return self.pool.set_queueprio(prio, self.queue)

    def is_started(self):
        return self.pool.is_started()

    def check_started(self):
        return self.pool.check_started()

    def stop(self, wait = False):
        # Must stop the main pool, not the wrapper
        # Don't complain though
        pass

    def close(self):
        pass

    def terminate(self):
        pass

    def start(self):
        return self.pool.start()

    def assert_started(self):
        return self.pool.assert_started()

    def join(self, timeout = None):
        # To-do: join only the subqueue
        return self.pool.join(timeout)

    def populate_workers(self):
        return self.pool.populate_workers()

    def apply_async(self, task, args = (), kwargs = {}):
        return self.pool.apply_async(task, args, kwargs, queue = self.queue)

    def apply(self, task, args = (), kwargs = {}, timeout = None):
        return self.pool.apply(task, args, kwargs, self.queue, timeout)




# -*- coding: utf-8 -*-
import os
import os.path

import ctypes
import mmap

import random
import tempfile
import threading
import operator

try:
    import fcntl
except ImportError:
    # Platform sucks
    fcntl = None # lint:ok

try:
    import numpy
except ImportError:
    try:
        # PyPy ?
        import numpypy as numpy  # lint:ok
    except ImportError:
        # Ouch
        numpy = None  # lint:ok

_implementations = []

class Slot(object):
    def __init__(self, slot_mmap, offset, dtype):
        self._dtype = dtype
        self._offset = offset
        self._mmap = slot_mmap

    @property
    def value(self):
        return self._mmap[self._offset]

    @value.setter
    def value(self, val):
        self._mmap[self._offset] = self._dtype(val or 0)

    @property
    def flat(self):
        return self._mmap

    def __iadd__(self, val):
        self._mmap[self._offset] += val

    def __isub__(self, val):
        self._mmap[self._offset] -= val

    def __add__(self, val):
        return self._mmap[self._offset] + val

    def __sub__(self, val):
        return self._mmap[self._offset] + val

    def __ne__(self, val):
        return self._mmap[self._offset] != val.value

    def __eq__(self, val):
        return self._mmap[self._offset] == val.value

    def __index__(self):
        return int(self.value)

class SlowSlot(object):
    def __init__(self, raw_mmap, raw_offset, slot_mmap, offset, dtype, ctype):
        self._dtype = dtype
        self._ctype = ctype
        self._offset = offset
        self._mmap = slot_mmap
        self._raw_slice = slice(raw_offset, raw_offset+ctypes.sizeof(ctype()))
        self._raw_mmap = raw_mmap

    @property
    def value(self):
        return self._mmap[self._offset]

    @value.setter
    def value(self, val):
        self._raw_mmap[self._raw_slice] = bytes(self._ctype(val or 0))

    @property
    def flat(self):
        return self._mmap

    def __iadd__(self, val):
        self.value = self._mmap[self._offset] + val

    def __isub__(self, val):
        self.value = self._mmap[self._offset] - val

    def __add__(self, val):
        return self._mmap[self._offset] + val

    def __sub__(self, val):
        return self._mmap[self._offset] + val

    def __ne__(self, val):
        return self._mmap[self._offset] != val.value

    def __eq__(self, val):
        return self._mmap[self._offset] == val.value

    def __index__(self):
        return int(self.value)


class SharedCounterGenericBase(object):
    """
    Maps size() bytes from the buffer into a shared counters
    structure with process-local subcounters. Reading the
    counters takes O(slots) time, updating takes
    O(1) and no operation blocks on other processes.
    Reads are only guaranteed to be eventually consistent,
    no snapshot semantics are guaranteed.

    If locked is given and True, slots must only be big
    enough to hold all concurrent process subcounters,
    and it is assumed the whole mapping operation will
    be free from concurrent mapping operations.

    If locked isn't given or is False, only a pid-derived
    counter can be used by this process, and if the slot
    has been used by another process, an AssertionError
    will be raised.

    SharedCounterGenericBase objects and derivations
    are both thread and process-safe, as long as you don't
    map a file/buffer more than once through different instances.
    """
    # Pick one
    dtype = None
    btype = None
    slots_item_size = None # make sizeof(dtype)
    bitmap_item_size = None # make sizeof(btype)

    def __init__(self, slots, bitmap, counters, locked, bitmap_slot):
        self.bitmap = None
        self.slots = None
        self.slot = None
        self.basemap = None
        self.baseoffset = None
        self.timestamp = None
        self.cached_timestamp = None
        self.cached_value = None

        # Try to acquire the slot
        slot = os.getpid() % slots
        if bitmap[slot]:
            if not locked:
                raise AssertionError("Slot occupied")
            else:
                # With a locked bitmap, we can search other slots
                for offs in range(slots):
                    nslot = (slot+offs) % slots
                    if not bitmap[nslot]:
                        slot = nslot
                        break
                else:
                    raise AssertionError("All slots occupied")

        self.bitmap = bitmap
        self.bitmap_slot = bitmap_slot(slot)
        bitmap_slot.value = True
        self.slots = counters
        self.slot = slot
        self.__rnd = random.getrandbits(62) + self.slot
        self.__wlock = threading.Lock()

    @classmethod
    def size(cls, slots):
        return (cls.slots_item_size + cls.bitmap_item_size) * slots + ctypes.sizeof(ctypes.c_uint64())

    def _update_ts(self):
        # Careful to maintain out-of-sync status out-of-sync
        ts = self.timestamp
        rand = self.__rnd
        ts.value = nts = (int(ts.value) + int(rand)) & 0xffffffffffffffff
        if self.cached_timestamp is not None:
            uint = type(self.cached_timestamp)
            self.cached_timestamp = uint(nts)

    @property
    def value(self):
        if self.cached_value is None or self.cached_timestamp != self.timestamp.value:
            self.cached_timestamp = self.timestamp.value
            self.cached_value = self._value
            self.__rnd = random.getrandbits(62) + self.slot
        return self.cached_value

    def flush(self):
        if self.basemap is not None:
            self.basemap.flush()

    @property
    def _value(self):
        # Really, provide a better implementation
        return sum(self.slots)

    @classmethod
    def from_path(cls, slots, pathname, offset = 0):
        """
        Maps the contents of fileo at offset "offset" interpreting
        them as shared counters of "slots" slots, locking the file
        for exclusive access during the duration of slot acquisition
        (which is quick) if a suitable locking implementation is present.

        If the file does not exist, it creates them in a manner that
        avoids race conditions, and initializes the counter to 0.

        Raises AssertionError if it cannot aquire the slot.
        """
        if not os.path.exists(pathname) or not os.access(pathname, os.R_OK | os.W_OK):
            # Initialize
            #
            # To make it portable, and atomic on all platforms with any filesystem,
            # we create a temporary unique file and atomically put it into our destination.
            #
            # We must make sure the temporary is on the same location as the destination,
            # to make sure the rename is quick and atomic.
            #
            # On unix-like systems, rename overwrites silently, so we use a link instead,
            # which doesn't. On other systems, we use rename.
            #
            # In any case, if the rename fails, we assume someone won over us, and open
            # the existing file instead.
            try:
                tmpfileno, tmppath = tempfile.mkstemp(
                    prefix = os.path.basename(pathname),
                    dir = os.path.dirname(pathname) )
                try:
                    # Fill with zeros
                    size = cls.size(slots)
                    zeros = b'\x00' * 1024
                    while size >= 1024:
                        os.write(tmpfileno, zeros)
                        size -= 1024
                    if size:
                        os.write(tmpfileno, zeros[:size])
                    # WHYNOT zeros = '\x00' * size; os.write(tmfileno, zeros)

                    # And swap
                    if hasattr(os, 'link'):
                        os.link(tmppath, pathname)
                        os.unlink(tmppath)
                    else:
                        os.rename(tmppath, pathname)
                    return cls.from_fileno(slots, tmpfileno, offset)
                finally:
                    os.close(tmpfileno)
            except OSError:
                # Let the fallback below try
                pass

        # We have a locking implementation, so it's as easy as just opening
        # the file and using it. The locks will take care of only using
        # an initialized counter file
        fileno = os.open(pathname, os.O_RDWR)
        try:
            return cls.from_fileno(slots, fileno, offset)
        finally:
            os.close(fileno)

    @classmethod
    def from_file(cls, slots, fileobj, offset = 0):
        """
        Maps the contents of fileobj at offset "offset" interpreting
        them as shared counters of "slots" slots, locking the file
        for exclusive access during the duration of slot acquisition
        (which is quick) if a suitable locking implementation is present.

        Raises AssertionError if it cannot aquire the slot.
        """
        return cls.from_fileno(slots, fileobj.fileno(), offset)

    @classmethod
    def from_fileno(cls, slots, fileno, offset = 0):
        """
        Maps the contents of the given file descriptor at offset "offset"
        interpreting them as shared counters of "slots" slots, locking the
        file for exclusive access during the duration of slot acquisition
        (which is quick) if a suitable locking implementation is present.

        Raises AssertionError if it cannot aquire the slot.
        """
        base_offset = offset - (offset % mmap.ALLOCATIONGRANULARITY)
        remnant_offset = offset - base_offset
        size = cls.size(slots)

        buf = mmap.mmap(fileno, size, offset = base_offset)

        rv = cls.from_buffer(slots, buf, remnant_offset, fileno, offset)
        rv.basemap = buf
        rv.baseoffset = offset
        return rv

    @classmethod
    def from_buffer(cls, slots, buf, offset = 0, fileno = None, fileoffs = None):
        """
        Maps the contents of buf interpreting
        it as shared counters of "slots" slots, locking the file descriptor
        given by fileno if not None, for exclusive access during the duration
        of slot acquisition (which is quick) if a suitable locking
        implementation is present.

        If given, fileoffs is the start of the region to be locked
        in fileno.

        Raises AssertionError if it cannot aquire the slot.
        """
        locked = fcntl is not None and fileno is not None
        size = cls.size(slots)
        try:
            if locked:
                fcntl.lockf(fileno, fcntl.LOCK_EX, size, fileoffs)
            return cls(slots, buf, offset, locked)
        finally:
            if locked:
                fcntl.lockf(fileno, fcntl.LOCK_UN, size, fileoffs)

    def __int__(self):
        return int(self.value)

    def __long__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __iadd__(self, val):
        with self.__wlock:
            self.myslot.value += val
            if self.cached_value is not None:
                self.cached_value += val
            self._update_ts()
        return self

    def __isub__(self, val):
        with self.__wlock:
            self.myslot.value -= val
            if self.cached_value is not None:
                self.cached_value -= val
            self._update_ts()
        return self

    def close(self):
        bitmap = getattr(self, 'bitmap', None)
        if bitmap is not None:
            # Release slot
            self.bitmap_slot.value = False
        del bitmap

        # Flush any pending writes
        basemap = getattr(self, 'basemap', None)
        if basemap is not None:
            basemap.flush()

        # Release possibly fd-holding resources
        self.slots = None
        self.slot = None
        self.basemap = None
        self.bitmap = None
        self.bitmap_slot = None
        self.timestamp = None
        self.cached_timestamp = None
        self.cached_value = None
        self.myslot = None

        if basemap is not None:
            basemap.close()

    def __del__(self):
        self.close()

    @property
    def _value(self):
        return self.slots.sum()

if numpy is not None:

    class SharedCounterBaseNumpyCtypes(SharedCounterGenericBase):
        btype = numpy.bool8
        bitmap_item_size = btype().itemsize

        @staticmethod
        def is_supported():
            if numpy is not None and getattr(numpy, 'frombuffer', None) is not None \
                    and hasattr(ctypes.c_uint64, 'from_buffer'):
                try:
                    amap = mmap.mmap(-1, 128)
                    bitmap = (ctypes.c_bool * 128).from_buffer(amap, 0)
                    bitmap[0] = 3
                    bitmap[5] = 3
                    return True
                except:
                    return False
            return False

        def __init__(self, slots, buf, offset = 0, locked = False):
            assert ctypes.sizeof(ctypes.c_bool()) == self.bitmap_item_size
            assert ctypes.sizeof(self.cdtype()) == self.slots_item_size

            timestamp = ctypes.c_uint64.from_buffer(buf, offset)
            offset += ctypes.sizeof(timestamp)

            # Slow, read-write bitmap
            bitmap = (ctypes.c_bool * slots).from_buffer(buf, offset)
            bitmap_slot = lambda slot : ctypes.c_bool.from_buffer(buf, offset + self.bitmap_item_size * slot)

            # Fast, read-only counters
            counters = numpy.frombuffer(buf, self.dtype, slots,
                offset + self.bitmap_item_size * slots)
            super(SharedCounterBaseNumpyCtypes, self).__init__(slots, bitmap, counters, locked, bitmap_slot)

            # Fast read-only bitmap ]:-]
            self.bitmap = numpy.frombuffer(buf, numpy.bool8, slots, offset)

            # Map my slot as a single item, it makes += atomic
            self.myslot = self.cdtype.from_buffer(buf,
                offset + ctypes.sizeof(bitmap) + self.slots_item_size * self.slot)

            # Map timestamp
            self.timestamp = timestamp

    class SharedCounter32NumpyCtypes(SharedCounterBaseNumpyCtypes):
        dtype = numpy.int32
        cdtype = ctypes.c_int32
        slots_item_size = dtype().itemsize
    class SharedCounter64NumpyCtypes(SharedCounterBaseNumpyCtypes):
        dtype = numpy.int64
        cdtype = ctypes.c_int64
        slots_item_size = dtype().itemsize


    class SharedCounterBaseNumpy(SharedCounterGenericBase):
        btype = numpy.bool8
        bitmap_item_size = btype().itemsize

        @staticmethod
        def is_supported():
            if numpy is not None and getattr(numpy, 'memmap', None) is not None:
                try:
                    import tempfile
                    with tempfile.NamedTemporaryFile() as temp:
                        ts_mmap = numpy.memmap(temp, numpy.uint64, 'r+', 0, 1)
                        ts_mmap[0] = 3
                    return True
                except:
                    return False
            else:
                return False

        def __init__(self, slots, fileobj, offset = 0, locked = False):
            ts_mmap = numpy.memmap(fileobj, numpy.uint64, 'r+', offset, 1)
            timestamp = Slot(ts_mmap, 0, numpy.uint64)
            offset += numpy.uint64().itemsize

            bitmap = numpy.memmap(fileobj, self.btype, 'r+', offset, slots)
            bitmap_slot = lambda slot : Slot(bitmap, slot, self.btype)
            offset += self.btype().itemsize * slots

            counters = numpy.memmap(fileobj, self.dtype, 'r+', offset, slots)
            offset += self.dtype().itemsize * slots
            super(SharedCounterBaseNumpy, self).__init__(slots, bitmap, counters, locked, bitmap_slot)

            self.myslot = Slot(counters, self.slot, self.dtype)

            self.timestamp = timestamp

        @classmethod
        def from_fileno(cls, slots, fileno, offset = 0):
            fd = os.dup(fileno)
            fileobj =  os.fdopen(fd, 'r+')
            return cls.from_file(slots, fileobj, offset)

        @classmethod
        def from_file(cls, slots, fileobj, offset = 0):
            base_offset = offset - (offset % mmap.ALLOCATIONGRANULARITY)
            remnant_offset = offset - base_offset
            size = cls.size(slots)

            rv = cls(slots, fileobj, offset)
            rv.basemap = mmap.mmap(fileobj.fileno(), size + remnant_offset, offset = base_offset)
            return rv

    class SharedCounter32Numpy(SharedCounterBaseNumpy):
        dtype = numpy.int32
        slots_item_size = numpy.dtype(dtype).itemsize
    class SharedCounter64Numpy(SharedCounterBaseNumpy):
        dtype = numpy.int64
        slots_item_size = numpy.dtype(dtype).itemsize


    class SharedCounterBaseReadOnlyNumpy(SharedCounterGenericBase):
        btype = numpy.bool8
        cbtype = ctypes.c_bool
        bitmap_item_size = btype().itemsize

        @staticmethod
        def is_supported():
            if numpy is not None and getattr(numpy, 'memmap', None) is not None:
                try:
                    import tempfile
                    with tempfile.NamedTemporaryFile() as temp:
                        ts_mmap = numpy.memmap(temp, numpy.uint64, 'r+', 0, 1)
                        raw_mmap = mmap.mmap(temp.fileno(), numpy.uint64().itemsize)
                        SlowSlot(raw_mmap, 0, ts_mmap, 0, numpy.uint64, ctypes.c_uint64).value = 3
                        ok = ts_mmap[0] == 3
                    return ok
                except:
                    return False
            else:
                return False

        def __init__(self, slots, fileobj, offset = 0, locked = False):
            assert ctypes.sizeof(ctypes.c_bool()) == self.bitmap_item_size
            assert ctypes.sizeof(self.cdtype()) == self.slots_item_size

            ts_mmap = numpy.memmap(fileobj, numpy.uint64, 'r+', offset, 1)
            if getattr(self, 'basemap', None) is not None:
                raw_mmap = self.basemap
            else:
                raw_mmap = self.basemap = mmap.mmap(fileobj.fileno(), self.size(slots), offset = offset)
            timestamp = SlowSlot(raw_mmap, offset, ts_mmap, 0, numpy.uint64, ctypes.c_uint64)
            offset += numpy.uint64().itemsize

            bitmap = numpy.memmap(fileobj, self.btype, 'r+', offset, slots)
            bitmap_slot = lambda slot : SlowSlot(raw_mmap, offset + slot, bitmap, slot, self.btype, self.cbtype)
            offset += self.btype().itemsize * slots

            counters = numpy.memmap(fileobj, self.dtype, 'r+', offset, slots)
            super(SharedCounterBaseReadOnlyNumpy, self).__init__(slots, bitmap, counters, locked,
                bitmap_slot = bitmap_slot)

            self.myslot = SlowSlot(raw_mmap, offset + self.dtype().itemsize * self.slot, counters,
                self.slot, self.dtype, self.cdtype)

            self.timestamp = timestamp

        @classmethod
        def from_fileno(cls, slots, fileno, offset = 0):
            fd = os.dup(fileno)
            fileobj =  os.fdopen(fd, 'r+')
            return cls.from_file(slots, fileobj, offset)

        @classmethod
        def from_file(cls, slots, fileobj, offset = 0):
            base_offset = offset - (offset % mmap.ALLOCATIONGRANULARITY)
            remnant_offset = offset - base_offset
            size = cls.size(slots)

            rv = cls(slots, fileobj, offset)
            rv.basemap = mmap.mmap(fileobj.fileno(), size + remnant_offset, offset = base_offset)
            return rv

    class SharedCounter32ReadOnlyNumpy(SharedCounterBaseReadOnlyNumpy):
        dtype = numpy.int32
        cdtype = ctypes.c_int32
        slots_item_size = numpy.dtype(dtype).itemsize
    class SharedCounter64ReadOnlyNumpy(SharedCounterBaseReadOnlyNumpy):
        dtype = numpy.int64
        cdtype = ctypes.c_int64
        slots_item_size = numpy.dtype(dtype).itemsize


class SharedCounterBaseCtypes(SharedCounterGenericBase):  # lint:ok
    btype = ctypes.c_bool
    bitmap_item_size = ctypes.sizeof(btype())

    @staticmethod
    def is_supported():
        if hasattr(ctypes.c_uint64, 'from_buffer'):
            try:
                amap = mmap.mmap(-1, 128)
                bitmap = (ctypes.c_bool * 128).from_buffer(amap, 0)
                bitmap[0] = 3
                bitmap[5] = 3
                return True
            except:
                return False
        return False

    def __init__(self, slots, buf, offset = 0, locked = False):
        timestamp = ctypes.c_uint64.from_buffer(buf, offset)
        offset += ctypes.sizeof(timestamp)

        bitmap = (ctypes.c_bool * slots).from_buffer(buf, offset)
        bitmap_slot = lambda slot : ctypes.c_bool.from_buffer(buf, offset + self.bitmap_item_size * slot)
        counters = (self.dtype * slots).from_buffer(buf, offset + ctypes.sizeof(bitmap))
        super(SharedCounterBaseCtypes, self).__init__(slots, bitmap, counters, locked, bitmap_slot)

        # Map my slot as a single item, it makes += atomic
        self.myslot = self.dtype.from_buffer(buf,
            offset + ctypes.sizeof(bitmap) + self.slots_item_size * self.slot)

        # Map timestamp
        self.timestamp = timestamp

    _implementations.extend([
        ( SharedCounterBaseNumpyCtypes, SharedCounter32NumpyCtypes, SharedCounter64NumpyCtypes ),
        ( SharedCounterBaseNumpy, SharedCounter32Numpy, SharedCounter64Numpy ),
    ])

class SharedCounter32Ctypes(SharedCounterBaseCtypes):
    dtype = ctypes.c_int32
    slots_item_size = ctypes.sizeof(dtype())
class SharedCounter64Ctypes(SharedCounterBaseCtypes):
    dtype = ctypes.c_int64
    slots_item_size = ctypes.sizeof(dtype())

_implementations.extend([
    ( SharedCounterBaseCtypes, SharedCounter32Ctypes, SharedCounter64Ctypes ),
])

if numpy is not None:
    # Last, low-priority implementation really
    _implementations.extend([
        ( SharedCounterBaseReadOnlyNumpy, SharedCounter32ReadOnlyNumpy, SharedCounter64ReadOnlyNumpy ),
    ])

for sbase, s32, s64 in _implementations:
    try:
        if sbase.is_supported():
            SharedCounterBase = sbase
            SharedCounter32 = s32
            SharedCounter64 = s64
            break
    except:
        pass
else:
    raise Exception(
        'No suitable shared counter implementaiton found. Try installing numpy and ensure support for '
        'read-write memmapped arrays'
    )
del sbase, s32, s64



from chorde.serialize import serialize

import logging
import time
import weakref
import functools
import itertools
import _thread as thread
import threading
import operator
import sys

# No need for real multiprocessing. In fact, using real
# multiprocessing would force pickling of values, which would be
# undesirable, pushing pickling cost into foreground threads.
import multiprocessing.dummy
import multiprocessing.pool
from chorde.threadpool import ThreadPool

from .base import BaseCacheClient, CacheMissError, NONE, TimeoutError, CancelledError
from .inproc import Cache

class _NONE:pass
class _DELETE:pass
class _EXPIRE:pass
class _PURGE:pass
class _CLEAR:pass
class _RENEW:pass
class REGET:pass

COALESCE_IGNORE_KWARGS = frozenset(['promote_callback'])
SPECIAL = set(map(id,[_NONE, _DELETE, _EXPIRE, _PURGE, _CLEAR, _RENEW, REGET]))
def is_special(value):
    return id(value) in SPECIAL


class Defer(object):
    """
    Wrap a callable in this, and pass it as a value to an AsyncWriteCacheClient,
    and the evaluation of the given callable will happen asynchronously. The cache
    will return stale entries until the callable is finished computing a new value.

    If a future attribute is attached, this defer will act as executioner of that
    future, and both set_running_or_notify_cancelled and set_result or set_exception
    will be called on it.

    If a future is attached during execution, set_running_or_notify_cancelled will
    not be invoked, but set_result will.
    """

    def __init__(self, callable_, *args, **kwargs):
        self.callable_ = callable_
        self.args = args
        self.kwargs = kwargs
        self.lazy = False

    def undefer(self, getattr=getattr):
        future = getattr(self, 'future', None)
        if future is None or future.set_running_or_notify_cancelled():
            try:
                self.rv = rv = self.callable_(*self.args, **self.kwargs)
            except:
                future = getattr(self, 'future', None)
                if future is not None:
                    future.exc(sys.exc_info())
                raise
            return rv
        else:
            future = getattr(self, 'future', None)
            if future is not None and not future.done():
                future.set_exception(CancelledError())
            return _NONE

    def set(self, value):
        self.rv = value

    def done(self, getattr=getattr):
        future = getattr(self, 'future', None)
        if future is not None and not future.done():
            rv = getattr(self, 'rv', NONE)
            if rv is _NONE or rv is NONE:
                future.set_exception(CancelledError())
            else:
                future.set(rv)

def _renew_undefer(self, key, ttl, kw, value):
    self = self()
    self.client.renew(key, ttl, **(kw or {}))
    return value.undefer()

_global_cleanup_tasks = []

class AsyncCacheWriterThreadPool(ThreadPool):
    class AsyncCacheWriterThread(ThreadPool.Process):
        pass
    Process = AsyncCacheWriterThread

    def __init__(self, workers):
        if ThreadPool is multiprocessing.pool.ThreadPool:
            # This patches ThreadPool, which is broken when instanced
            # from inside a DummyThread (happens after forking)
            current = multiprocessing.dummy.current_process()
            if not hasattr(current, '_children'):
                current._children = weakref.WeakKeyDictionary()

        ThreadPool.__init__(self, workers)

class AsyncCacheWriterPool:
    def __init__(self, size, workers, client, overflow = False, cleanup_cycles = 500,
            defer_threadpool = None, writer_threadpool = None):

        self.client = client
        self.logger = logging.getLogger("chorde.async")
        self.size = size
        self.workers = workers
        self._spawnlock = threading.Lock()
        if callable(defer_threadpool):
            self._defer_threadpool_factory = defer_threadpool
            self._defer_threadpool = None
        else:
            self._defer_threadpool_factory = None
            self._defer_threadpool = defer_threadpool
        if callable(writer_threadpool):
            self._writer_threadpool_factory = writer_threadpool
            self._writer_threadpool = None
        else:
            self._writer_threadpool_factory = None
            self._writer_threadpool = writer_threadpool

        # queueset holds the values to be written, associated
        # by key, providing some write-back coalescense in
        # high-load environments
        self.queueset = {}
        self.workset = {}
        self.threadset = set()
        self.done_event = threading.Event()
        self.overflow = overflow
        self._wself = weakref.ref(self)

        self.tl = threading.local()
        self.cleanup_tasks = []
        self.cleanup_cycles = cleanup_cycles

    @property
    def defer_threadpool(self):
        if self._defer_threadpool is None:
            with self._spawnlock:
                if self._defer_threadpool is None:
                    if self._defer_threadpool_factory is not None:
                        self._defer_threadpool = self._defer_threadpool_factory(self.workers)
                    else:
                        self._defer_threadpool = AsyncCacheWriterThreadPool(self.workers)
        return self._defer_threadpool

    @property
    def writer_threadpool(self):
        if self._writer_threadpool is None:
            with self._spawnlock:
                if self._writer_threadpool is None:
                    if self._writer_threadpool_factory is not None:
                        self._writer_threadpool = self._writer_threadpool_factory(self.workers)
                    else:
                        self._writer_threadpool = AsyncCacheWriterThreadPool(self.workers)
        return self._writer_threadpool

    @staticmethod
    def _writer(self, key, reentrant = True):
        # self is weakref
        self = self()
        if self is None:
            return

        if reentrant:
            thread_id = thread.get_ident()
            if thread_id not in self.threadset:
                self.threadset.add(thread_id)
            else:
                thread_id = None
        else:
            thread_id = None

        ev = self.done_event
        deferred = _NONE
        value = self.dequeue(key)
        try:
            if value is _NONE or value is NONE:
                # Cancelled
                return
            else:
                # Unpack
                value, ttl, kw = value

            if value is _NONE or value is NONE:
                # Something's hinky
                return
            elif hasattr(value, 'undefer'):
                deferred = value
                try:
                    value = value.undefer()
                    if value is REGET:
                        deferred.set(self.client.get(key))
                        value = _NONE # don't set, it's already set
                except CacheMissError:
                    # It's ok, accepted pattern to cancel computation in a transparent way
                    value = _NONE
                except:
                    self.logger.error("Error in background cache refresh", exc_info=True)
                    value = _NONE

            if value is _NONE or value is NONE:
                # undefer probably decided not to compute anything (or an error arose, whatever)
                if deferred is not _NONE:
                    deferred.done()
                return

            elif value is _DELETE:
                try:
                    self.client.delete(key)
                except:
                    self.logger.error("Error deleting key", exc_info=True)

            elif value is _RENEW:
                try:
                    self.client.renew(key, ttl, **(kw or {}))
                except:
                    self.logger.error("Error renewing key", exc_info=True)

            elif value is _EXPIRE:
                try:
                    self.client.expire(key)
                except:
                    self.logger.error("Error expiring key", exc_info=True)

            elif value is _CLEAR:
                try:
                    self.client.clear()
                except:
                    self.logger.error("Error clearing cache", exc_info = True)

            elif value is _PURGE:
                try:
                    self.client.purge()
                except:
                    self.logger.error("Error purging cache", exc_info = True)

            else:
                try:
                    self.client.put(key, value, ttl, **(kw or {}))
                except:
                    self.logger.error("Error saving data in cache %r, key %r", self.client, key, exc_info=True)

            if deferred is not _NONE:
                deferred.done()
        finally:
            # Signal waiting threads
            w = self.workset.pop(key, None)
            if w is not None:
                kev = w[2]
            else:
                kev = None
            del w

            if thread_id is not None and thread_id not in map(operator.itemgetter(0), list(self.workset.values())):
                try:
                    self.threadset.remove(thread_id)
                except KeyError:
                    pass

            ev.set()
            if kev is not None:
                kev.set()

        # Cleanup
        try:
            if not hasattr(self.tl, 'dirty_rounds'):
                self.tl.dirty_rounds = 0
            self.tl.dirty_rounds += 1
            if self.tl.dirty_rounds > self.cleanup_cycles or not self.queueset:
                self.tl.dirty_rounds = 0
                for task in self.cleanup_tasks:
                    task()
                for task in _global_cleanup_tasks:
                    task()
        except:
            self.logger.error("Error during cleanup", exc_info = True)

    @property
    def capacity(self):
        return self.size

    @property
    def usage(self):
        return len(self.queueset)

    @serialize
    def dequeue(self, key):
        rv = self.queueset.pop(key, _NONE)
        self.workset[key] = thread.get_ident(), rv, threading.Event()
        return rv

    @serialize
    def drop_one(self):
        for key in self.queueset:
            break
        else:
            return
        rv = self.queueset.pop(key, _NONE)
        return rv

    def enqueue(self, key, value, ttl=None, **kw):
        if (thread.get_ident() in self.threadset
                 or (hasattr(self.defer_threadpool, 'in_worker') and self.defer_threadpool.in_worker()) ):
            # Oops, recursive call, bad idea
            # Run inline
            queueset = self.queueset
            if key in queueset:
                delayed = self._coalesce(key, value, ttl, **kw)
                if delayed is not None:
                    # delayed callback, invoke now that we're outside the critical section
                    delayed()
                else:
                    self._writer(self._wself, key)
            else:
                self.queueset[key] = value, ttl, (kw or None)
                self._writer(self._wself, key)
        else:
            if key not in self.queueset:
                if self.overflow:
                    # use overflow semantics: remove old entries to make room for new ones
                    if len(self.queueset) >= self.size:
                        # just two, tit-for-tat, one in, two out. Avoids large latencies,
                        # and guarantees stable sizes, around, while not strictly below "size"
                        for _ in range(2):
                            self.drop_one()
                            if len(self.queueset) < self.size:
                                break
                elif not is_special(value):
                    # blocking semantics, wait
                    ev = self.done_event
                    while len(self.queueset) >= self.size:
                        ev.wait(1.0)
                        if ev.isSet():
                            ev.clear()
            delayed = self._enqueue(key, value, ttl, **kw)
            if delayed is not None:
                # delayed callback, invoke now that we're outside the critical section
                delayed()

    @serialize
    def clearqueue(self):
        delayed = []
        for entry in self.queueset.values():
            value = entry[0]
            if hasattr(value, 'undefer') and hasattr(value, 'future'):
                future = getattr(value, 'future', None)
                if future is not None and hasattr(future, 'add_done_callback'):
                    delayed.append(future.cancel)
        self.queueset.clear()
        return delayed

    def _coalesce(self, key, value, ttl, **kw):
        delayed = None
        queueset = self.queueset
        if hasattr(value, 'undefer'):
            # Queued one wins, we just have to chain the futures if any
            future = getattr(value, 'future', None)
            if future is not None and hasattr(future, 'add_done_callback'):
                queue_value = queueset.get(key)
                if queue_value is not None:
                    # Ok, they'll wanna get the value when it's done
                    queue_value = queue_value[0]
                    if hasattr(queue_value, 'undefer'):
                        queue_future = getattr(queue_value, 'future', None)
                        if queue_future is not None:
                            delayed = functools.partial(queue_future.chain, future)
                        else:
                            queue_value.future = future
                    else:
                        # Why not
                        delayed = functools.partial(future.set, queue_value)
            else:
                queue_value = queueset.get(key)
                if queue_value is not None:
                    queue_value, queue_ttl, queue_kwargs = queue_value
                    if queue_value is _RENEW:
                        # Turn renew + put serial operations into a single renew-then-put defer
                        queueset[key] = (
                            Defer(_renew_undefer, self._wself, key, queue_ttl, queue_kwargs, value),
                            ttl,
                            (kw or None)
                        )
        else:
            # New one wins, we just have to chain any queued future
            queue_value = queueset.get(key)
            if queue_value is not None:
                # Ok, they'll wanna get the value when it's done
                queue_value = queue_value[0]
                if hasattr(queue_value, 'undefer'):
                    queue_future = getattr(queue_value, 'future', None)
                    if queue_future is not None:
                        # Delay the callback, we're in a critical section here
                        delayed = functools.partial(queue_future.set, value)
            queueset[key] = value, ttl, (kw or None)
        return delayed

    @serialize
    def _enqueue(self, key, value, ttl, isinstance=isinstance, getattr=getattr, hasattr=hasattr, **kw):
        delayed = None
        queueset = self.queueset
        workset = self.workset
        if key not in queueset:
            if not hasattr(value, 'undefer') or key not in workset:
                if hasattr(value, 'undefer'):
                    threadpool = self.defer_threadpool
                    reentrant = True
                else:
                    threadpool = self.writer_threadpool
                    reentrant = False
                queueset[key] = value, ttl, (kw or None)
                threadpool.apply_async(self._writer, (self._wself, key, reentrant))
            else:
                # else, bad luck, we assume defers compute, so if two
                # defers go in concurrently, only the first will be invoked,
                # instead of the last - the first cannot be canceled after all,
                # and we want to only invoke one. So no choice.

                # ...if the defer has a future attached
                future = getattr(value, 'future', None)
                if future is not None and hasattr(future, 'add_done_callback'):
                    # we do have to hook into the future though...
                    working = workset.get(key)
                    if working is not None:
                        working = working[1]
                        if working is not _NONE:
                            working = working[0]
                        if hasattr(working, 'undefer'):
                            working_future = getattr(working, 'future', None)
                            if working_future is not None and hasattr(working_future, 'chain'):
                                delayed = functools.partial(working_future.chain, future)
                            elif getattr(working, 'rv', _NONE) is not _NONE:
                                # Delay the callback, we're in a critical section here
                                delayed = functools.partial(future.set, working.rv)
                            else:
                                working.future = future
                        else:
                            # Delay the callback, we're in a critical section here
                            delayed = functools.partial(future.set, working)
        else:
            delayed = self._coalesce(key, value, ttl, **kw)
        return delayed

    def waitkey(self, key, timeout=None):
        tid = thread.get_ident()
        if (tid in self.threadset
                or (hasattr(self.defer_threadpool, 'in_worker') and self.defer_threadpool.in_worker())):
            # Oops, recursive call, bad idea
            return

        tidt = (tid,)
        queueset = self.queueset
        workset_get = self.workset.get
        done_event = self.done_event
        sleep = time.sleep
        busyloop = 0

        if timeout is None:
            # contains inlined for speed in code below
            while (key in queueset or workset_get(key, tidt)[0] != tid):
                ev = workset_get(key)
                if ev is not None:
                    ev = ev[2]
                elif (key in queueset or workset_get(key, tidt)[0] != tid):
                    ev = done_event
                if ev is not None:
                    done = ev.wait(1.0)
                    if not done:
                        busyloop = 0
                    else:
                        busyloop += 1
                        if busyloop > 5:
                            if ev is done_event:
                                if busyloop > 20 and ev.isSet():
                                    ev.clear()
                            else:
                                sleep(0.05)
                else:
                    break
        else:
            # contains inlined for speed in code below
            time_time = time.time
            tfin = time_time() + timeout
            while (key in queueset or workset_get(key, tidt)[0] != tid) and tfin >= time_time():
                ev = workset_get(key)
                if ev is not None:
                    ev = ev[2]
                elif (key in queueset or workset_get(key, tidt)[0] != tid):
                    ev = done_event
                if ev is not None:
                    done = ev.wait(min(1.0, timeout))
                    if not done:
                        busyloop = 0
                    else:
                        busyloop += 1
                        if busyloop > 5:
                            if ev is done_event:
                                if busyloop > 20 and ev.isSet():
                                    ev.clear()
                            else:
                                sleep(0.05)
                else:
                    break
                timeout = tfin - time_time()

    def getTtl(self, key, default = None):
        # Speeding up things - we have the same signature and semantics
        self.getTtl = getTtl = self.queueset.get
        return getTtl(key, default)

    def contains(self, key):
        # Not atomic, but it doesn't really matter much, very unlikely and benignly to fail
        if key in self.queueset:
            return True
        else:
            # Exclude keys being computed by this thread, such calls
            # want the async wrapper out of their way
            tid = thread.get_ident()
            return self.workset.get(key, (tid,))[0] != tid

    def _contains(self, key):
        # Fast version of contains that doesn't check the workset
        # Useful for special-purpose checks, like contains(CLEAR)
        self._contains = _contains = self.queueset.__contains__
        return _contains(key)

    def put(self, key, value, ttl, **kw):
        self.enqueue(key, value, ttl, **kw)

    def renew(self, key, ttl, **kw):
        # Don't schedule a renew if another thing is queued on the key
        # It causes... issues
        if key not in self.queueset and key not in self.workset:
            self.enqueue(key, _RENEW, ttl, **kw)

    def delete(self, key):
        self.enqueue(key, _DELETE)

    def expire(self, key):
        self.enqueue(key, _EXPIRE)

    def clear(self):
        delayed = self.clearqueue()
        self.enqueue(_CLEAR, _CLEAR)
        for delayed in delayed:
            try:
                delayed()
            except:
                pass

    def purge(self):
        self.enqueue(_PURGE, _PURGE)

    def register_cleanup(self, task):
        """
        Register a callable that will take no arguments, and will
        be invoked every cleanup_cycles tasks in order to perform
        thread-local cleanup for this writer.
        """
        self.cleanup_tasks.append(task)

    @staticmethod
    def register_default_cleanup(task):
        """
        Register a callable that will take no arguments, and will
        be invoked every cleanup_cycles tasks in order to perform
        thread-local cleanup for all async processors (including
        async writers and async processors both).
        """
        _global_cleanup_tasks.append(task)

class AsyncWriteCacheClient(BaseCacheClient):
    def __init__(self, client, writer_queue_size, writer_workers = None, overflow = False, threadpool = None,
            defer_threadpool = None, writer_threadpool = None):
        self.client = client
        self.writer_queue_size = writer_queue_size
        self.writer_workers = writer_workers if writer_workers is not None else multiprocessing.cpu_count()
        self.writer = None
        self.overflow = overflow
        self.spawning_lock = threading.Lock()
        self._defer_threadpool = defer_threadpool or threadpool
        self._writer_threadpool = writer_threadpool or threadpool

    def assert_started(self):
        if self.writer is None:
            with self.spawning_lock:
                if self.writer is None:
                    self.writer = AsyncCacheWriterPool(
                        self.writer_queue_size,
                        self.writer_workers,
                        self.client,
                        self.overflow,
                        defer_threadpool = self._defer_threadpool,
                        writer_threadpool = self._writer_threadpool)

    def is_started(self):
        return self.writer is not None

    def start(self):
        if self.writer is not None:
            raise AssertionError("Starting AsyncCacheClient twice")
        self.assert_started()

    def stop(self, abort_tasks=False):
        if self.writer is not None:
            if not abort_tasks:
                self.writer.join()
            self.writer.terminate()

    @property
    def is_async(self):
        return True

    @property
    def capacity(self):
        return (self.client.capacity, self.writer.capacity if self.writer is not None else 0)

    @property
    def usage(self):
        return (self.client.usage, self.writer.usage if self.writer is not None else 0)

    def put(self, key, value, ttl, **kw):
        self.assert_started()
        self.writer.put(key, value, ttl, **kw)

    def renew(self, key, ttl, **kw):
        self.assert_started()
        self.writer.renew(key, ttl, **kw)

    def delete(self, key):
        self.assert_started()
        self.writer.delete(key)

    def expire(self, key):
        self.assert_started()
        self.writer.expire(key)

    def clear(self):
        if self.is_started():
            self.writer.clear()

    def purge(self, timeout = None):
        """
        Timeout is ignored
        """
        if self.is_started():
            self.writer.purge()

    def getTtl(self, key, default = NONE,
            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
            NONE = NONE, _NONE = _NONE,
            hasattr = hasattr,
            **kw):
        ettl = None
        writer = self.writer
        if writer is not None: # self.is_started() inlined for speed
            # Try to read pending writes as if they were on the cache
            value = writer.getTtl(key, _NONE)
            if value is not _NONE:
                value, ttl, _ = value
                if value is _DELETE:
                    # Deletion means a miss... right?
                    if default is NONE:
                        raise CacheMissError(key)
                    else:
                        return default, -1
                elif value is _EXPIRE:
                    # Expiration just sets the TTL
                    ettl = -1
                elif value is _RENEW:
                    ettl = ttl
                elif not hasattr(value, 'undefer'):
                    return value, ttl
            # Yep, _NONE when querying the writer, because we don't want
            # to return a default if the writer doesn't have it, we must
            # still check the client.

            # Check pending clear - after checking the queue for sorted semantics
            if writer._contains(_CLEAR):
                # Well,
                if default is NONE:
                    raise CacheMissError(key)
                else:
                    return default, -1

        # Ok, read the cache then
        value, ttl = self.client.getTtl(key, default, **kw)
        if ettl is not None:
            ttl = ettl
        if value is NONE:
            raise CacheMissError(key)
        else:
            return value, ttl

    def get(self, key, default = NONE, ttl_skip = None,
            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
            NONE = NONE, _NONE = _NONE,
            hasattr = hasattr,
            **kw):
        writer = self.writer
        if writer is not None: # self.is_started() inlined for speed
            # Try to read pending writes as if they were on the cache
            value = writer.getTtl(key, _NONE)
            if value is not _NONE:
                value, ttl, _ = value
                if value is _DELETE or value is _EXPIRE:
                    # Deletion means a miss... right?
                    if default is NONE:
                        raise CacheMissError(key)
                    else:
                        return default
                elif value is _RENEW and (ttl_skip is None or ttl >= ttl_skip):
                    # The value is renewed, ttl_skip will be meaningless
                    ttl_skip = None
                elif not hasattr(value, 'undefer'):
                    return value
            # Yep, _NONE when querying the writer, because we don't want
            # to return a default if the writer doesn't have it, we must
            # still check the client.

            # Check pending clear - after checking the queue for sorted semantics
            if writer._contains(_CLEAR):
                # Well,
                if default is NONE:
                    raise CacheMissError(key)
                else:
                    return default

        # Ok, read the cache then
        return self.client.get(key, default, ttl_skip = ttl_skip, **kw)

    def getTtlMulti(self, keys, default = NONE,
            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
            NONE = NONE, _NONE = _NONE,
            hasattr = hasattr,
            **kw):
        ettl = {}
        writer = self.writer
        default_rv = (default, -1)
        if writer is not None: # self.is_started() inlined for speed
            # Try to read pending writes as if they were on the cache
            wgetTtl = writer.getTtl
            nkeys = []
            nkeys_append = nkeys.append
            for key in keys:
                value = wgetTtl(key, _NONE)
                if value is not _NONE:
                    value, ttl, _ = value
                    if value is _DELETE:
                        # Deletion means a miss... right?
                        yield key, default_rv
                        continue
                    elif value is _EXPIRE:
                        # Expiration just sets the TTL
                        ettl[key] = -1
                    elif value is _RENEW:
                        ettl[key] = ttl
                    elif not hasattr(value, 'undefer'):
                        yield key, (value, ttl)
                        continue
                # Yep, _NONE when querying the writer, because we don't want
                # to return a default if the writer doesn't have it, we must
                # still check the client.
                nkeys_append(key)
            keys = nkeys
            del nkeys, nkeys_append

            # Check pending clear - after checking the queue for sorted semantics
            if writer._contains(_CLEAR):
                for key in keys:
                    yield key, default_rv
                return

        # Ok, read the cache then
        ettlget = ettl.get
        for key, (value, ttl) in self.client.getTtlMulti(keys, default, **kw):
            ttl = ettlget(key, ttl)
            yield key, (value, ttl)

    def getMulti(self, keys, default = NONE, ttl_skip = None,
            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
            NONE = NONE, _NONE = _NONE,
            hasattr = hasattr,
            **kw):
        ettl = []
        ettl_append = ettl.append
        writer = self.writer
        if writer is not None: # self.is_started() inlined for speed
            # Try to read pending writes as if they were on the cache
            wgetTtl = writer.getTtl
            nkeys = []
            nkeys_append = nkeys.append
            for key in keys:
                value = wgetTtl(key, _NONE)
                if value is not _NONE:
                    value, ttl, _ = value
                    if value is _DELETE or value is _EXPIRE:
                        # Deletion/expiration means a miss...
                        yield key, default
                        continue
                    elif value is _RENEW and (ttl_skip is None or ttl >= ttl_skip):
                        if ttl_skip is not None:
                            # We'll do these without ttl_skip
                            ettl_append(key)
                            continue
                    elif not hasattr(value, 'undefer'):
                        yield key, value
                        continue
                # Yep, _NONE when querying the writer, because we don't want
                # to return a default if the writer doesn't have it, we must
                # still check the client.
                nkeys_append(key)
            keys = nkeys
            del nkeys, nkeys_append

            # Check pending clear - after checking the queue for sorted semantics
            if writer._contains(_CLEAR):
                for key in keys:
                    yield key, default
                for key in ettl:
                    yield key, default
                return

        # Ok, read the cache then
        if ettl:
            for key, value in self.client.getMulti(keys, default, **kw):
                yield key, value
        if keys:
            for key, value in self.client.getMulti(keys, default, ttl_skip = ttl_skip, **kw):
                yield key, value

    def promote(self, key, *p, **kw):
        if self.is_started() and self.writer.contains(key):
            return
        else:
            return self.client.promote(key, *p, **kw)

    def wait(self, key, timeout = None):
        if self.is_started() and self.writer.contains(key):
            self.writer.waitkey(key, timeout)

    def contains(self, key, ttl = None, **kw):
        if self.is_started():
            if self.writer.contains(key):
                return True
            else:
                return self.client.contains(key, ttl, **kw)
        else:
            return self.client.contains(key, ttl, **kw)


    def register_cleanup(self, task):
        """
        Register a callable that will take no arguments, and will
        be invoked every cleanup_cycles tasks in order to perform
        thread-local cleanup for this writer.
        """
        self.writer.register_cleanup(task)

    def __str__(self):
        return "<%s of %r>" % (self.__class__.__name__, self.client)

    def __repr__(self):
        return str(self)

try:
    from ._async import ExceptionWrapper, Future, set_strip_tracebacks
except ImportError:
    import warnings
    warnings.warn("_async extension module not built in, "
        "using pure-python version which is not atomic and requires"
        "explicit synchronization. Decreased performance will be noticeable")
    del warnings

    strip_tracebacks = False

    def set_strip_tracebacks(strip):
        """ Sets traceback stripping behavior for async futures

        Async tracebacks have a tendency to cause memory leaks. Stripping them in
        production deployments is a good idea to avoid them, but keep them around
        during testing or non-production environments to get better diagnostic
        information and be able to detect those leaks.
        """
        global strip_tracebacks
        strip_tracebacks = strip

    class ExceptionWrapper(object):  # lint:ok
        __slots__ = ('value',)

        def __init__(self, value):
            self.value = value

        def reraise(self, strip=True):
            exc = self.value
            if strip:
                del self.value
            try:
                exc_typ, exc_obj, exc_tb = exc
            finally:
                # Don't leave references to the exc/tb in the frame
                del exc
            try:
                if not strip:
                    # Can't raise the same exception object multiple times,
                    # tracebacks accumulate and leak. Use proper chaining.
                    raise exc_typ(*exc_obj.args) from exc_obj
                elif exc_tb is not None:
                    if exc_obj is not None:
                        if getattr(exc_obj, '__traceback__') is not exc_tb:
                            exc_obj = exc_obj.with_traceback(exc_tb)
                        raise exc_obj
                    else:
                        raise exc_typ().with_traceback(exc_tb)
                elif exc_obj is not None:
                    raise exc_obj
                else:
                    raise exc_typ()
            finally:
                # Don't leave references to the exc/tb in the frame
                del exc_typ, exc_obj, exc_tb

    class Future(object):  # lint:ok
        __slots__ = (
            '_cb', '_value', '_logger', '_running', '_cancel_pending', '_cancelled', '_done_event',
            '_lock', '__weakref__',
        )

        def __init__(self, logger = None):
            self._cb = []
            self._logger = logger
            self._lock = threading.Lock()

        def _set_nothreads(self, value, hasattr = hasattr, tuple = tuple, getattr = getattr):
            """
            Like set(), but assuming no threading is involved. It won't wake waiting threads,
            nor will it try to be thread-safe. Safe to call when the calling
            thread is the only one owning references to this future, and much faster.
            """
            if hasattr(self, '_value'):
                # No setting twice
                return

            self._value = value
            self._running = False

            if self._cb:
                cbs = list(self._cb)
                del self._cb[:len(cbs)]
                for cb in cbs:
                    try:
                        cb(value)
                    except:
                        if self._logger is not None:
                            error = self._logger
                        else:
                            error = logging.error
                        error("Error in async callback", exc_info = True)

        def set(self, value, hasattr = hasattr, tuple = tuple, getattr = getattr):
            """
            Set the future's result as either a value, an exception wrappedn in ExceptionWrapper, or
            a cache miss if given CacheMissError (the class itself)
            """
            if hasattr(self, '_value'):
                # No setting twice
                return

            with self._lock:
                old = getattr(self, '_value', None) # avoid deadlocks due to finalizers
                cbs = list(self._cb)
                self._value = value
            del old

            for cb in cbs:
                try:
                    cb(value)
                except:
                    if self._logger is not None:
                        error = self._logger
                    else:
                        error = logging.error
                    error("Error in async callback", exc_info = True)
            self._running = False

            event = getattr(self, '_done_event', None)
            if event is not None:
                # wake up waiting threads
                event.set()

        set_result = set

        def miss(self):
            """
            Shorthand for setting a cache miss result
            """
            self.set(CacheMissError)

        def _miss_nothreads(self):
            """
            Shorthand for setting a cache miss result without thread safety.
            See _set_nothreads
            """
            self._set_nothreads(CacheMissError)

        def exc(self, exc_info):
            """
            Shorthand for setting an exception result from an exc_info tuple
            as returned by sys.exc_info()
            """
            self.set(ExceptionWrapper(exc_info))

        def _exc_nothreads(self, exc_info):
            """
            Shorthand for setting an exception result from an exc_info tuple
            as returned by sys.exc_info(), without thread safety.
            See _set_nothreads
            """
            self._set_nothreads(ExceptionWrapper(exc_info))

        def set_exception(self, exception):
            """
            Set the Future's exception object.
            """
            if strip_tracebacks and getattr(exception, '__traceback__', None) is not None:
                exception = exception.with_traceback(None)
            self.exc((type(exception),exception,None))

        def on_value(self, callback):
            """
            When and if the operation completes without exception, the callback
            will be invoked with its result.
            """
            def value_callback(value):
                if value is not CacheMissError and not isinstance(value, ExceptionWrapper):
                    return callback(value)
            return self._on_stuff(value_callback)

        def on_miss(self, callback):
            """
            If the operation results in a cache miss, the callback will be invoked
            without arugments.
            """
            def miss_callback(value):
                if value is CacheMissError:
                    return callback()
            return self._on_stuff(miss_callback)

        def on_exc(self, callback):
            """
            If the operation results in an exception, the callback will be invoked
            with an exc_info tuple as returned by sys.exc_info.
            """
            def exc_callback(value):
                if isinstance(value, ExceptionWrapper):
                    return callback(value.value)
            return self._on_stuff(exc_callback)

        def on_any(self, on_value = None, on_miss = None, on_exc = None):
            """
            Handy method to set callbacks for all kinds of results, and it's actually
            faster than calling on_X repeatedly. None callbacks will be ignored.
            """
            def callback(value):
                if value is CacheMissError:
                    if on_miss is not None:
                        return on_miss()
                elif isinstance(value, ExceptionWrapper):
                    if on_exc is not None:
                        return on_exc(value.value)
                else:
                    if on_value is not None:
                        return on_value(value)
            return self._on_stuff(callback)

        def on_any_once(self, on_value = None, on_miss = None, on_exc = None):
            """
            Like on_any, but will only set the callback if no other callback has been set
            """
            if not self._cb:
                self.on_any(on_value, on_miss, on_exc)

        def on_done(self, callback):
            """
            When the operation is done, the callback will be invoked without arguments,
            regardless of the outcome. If the operation is cancelled, it won't be invoked.
            """
            def done_callback(value):
                return callback()
            return self._on_stuff(done_callback)

        def chain(self, defer):
            """
            Invoke all the callbacks of the other defer
            """
            self._on_stuff(defer.set)

        def chain_std(self, defer):
            """
            Invoke all the callbacks of the other defer, without assuming the other
            defer follows our non-standard interface.
            """
            return self.on_any(
                defer.set_result,
                functools.partial(defer.set_exception, CacheMissError()),
                lambda value : defer.set_exception(value[1] or value[0])
            )

        def _on_stuff(self, callback, hasattr=hasattr):
            cbap = self._cb.append
            docall = hasattr(self, '_value')
            if not docall:
                with self._lock:
                    docall = hasattr(self, '_value')
                    if not docall:
                        cbap(callback)
            if docall:
                callback(self._value)
            return self

        def add_done_callback(self, callback):
            """
            When the operatio is done, the callback will be invoked with the
            future object as argument.
            """
            me = weakref.ref(self)
            def weak_callback(value):
                self = me()
                if self is not None:
                    return callback(self)
            return self._on_stuff(weak_callback)

        def done(self, hasattr=hasattr, getattr=getattr):
            """
            Return True if the operation has finished, in a result or exception or cancelled, and False if not.
            """
            return hasattr(self, '_value') or getattr(self, '_cancelled', False)

        def running(self, getattr=getattr):
            """
            Return True if the operation is running and cannot be cancelled. False if not running
            (yet or done).
            """
            return getattr(self, '_running', False)

        def cancelled(self, getattr=getattr):
            """
            Return True if the operation has been cancelled successfully.
            """
            return getattr(self, '_cancelled', False)

        def cancel_pending(self, getattr=getattr):
            """
            Return True if cancel was called.
            """
            return getattr(self, '_cancel_pending', False)

        def cancel(self, getattr=getattr):
            """
            Request cancelling of the operation. If the operation cannot be cancelled,
            it will return False. Otherwise, it will return True.
            """
            if getattr(self, '_cancelled', False):
                return False
            else:
                self._cancel_pending = True
                return True

        def set_running_or_notify_cancelled(self, getattr=getattr):
            """
            To be invoked by executors before executing the operation. If it returns True,
            the operation may go ahead, and if False, a cancel has been requested and the
            operation should not be initiated, all threads waiting for the operation will
            be wakened immediately and the future will be marked as cancelled.
            """
            if getattr(self, '_cancel_pending', False):
                self._cancelled = True
                self._running = False

                # Notify waiters and callbacks
                self.set_exception(CancelledError())

                return False
            else:
                self._running = True
                return True

        def result(self, timeout=None, hasattr=hasattr, getattr=getattr, isinstance=isinstance, norecurse=False):
            """
            Return the operation's result, if any. If an exception was the result, re-raise it.
            If it was cancelled, raises CancelledError, and if timeout is specified and not None,
            and the specified time elapses without a result available, raises TimeoutError.
            """
            if hasattr(self, '_value'):
                value = self._value
                if isinstance(value, ExceptionWrapper):
                    try:
                        value.reraise(False)
                    finally:
                        del value
                elif value is CacheMissError:
                    raise CacheMissError()
                else:
                    return self._value
            elif self.cancelled():
                raise CancelledError()
            else:
                if timeout == 0:
                    raise TimeoutError()
                else:
                    # Wait for it
                    event = getattr(self, '_done_event', None)
                    if event is None:
                        event = self._done_event = threading.Event()
                    # First loop eagerly waits on the recently-created event
                    # Second loop grabs the instance event (which could have been
                    # clobbered by another thread). This is lockless yet safe,
                    # and quick on the most common condition (no contention)
                    for timeout in (0, timeout):
                        if event.wait(timeout) and not norecurse:
                            return self.result(0, norecurse=True)
                        elif self.cancelled():
                            raise CancelledError()
                        else:
                            time.sleep(0) # < give other threads a chance
                            event = self._done_event
                    else:
                        raise TimeoutError()

        def exception(self, timeout=None):
            """
            If the operation resulted in an exception, return the exception object.
            Otherwise, return None. If the operation has been cancelled, raises CancelledError,
            and if timeout is specified and not None, and the specified time elapses without
            a result available, raises TimeoutError.
            """
            if hasattr(self, '_value'):
                value = self._value
                if isinstance(value, ExceptionWrapper):
                    return value.value[1] or value.value[0]
                elif value is CacheMissError:
                    return CacheMissError
                else:
                    return None
            elif self.cancelled():
                raise CancelledError()
            else:
                try:
                    self.result()
                    return None
                except CancelledError:
                    raise
                except Exception as e:
                    try:
                        return e
                    finally:
                        del e


def makeFutureWrapper(base):
    """
    Nice utility function to create Future wrappers. If using a library
    with concurrent-compatible futures, but not quite the same, and the library
    checks for inheritance instead of relying on duck typing, then you'll
    need to make such a class and wrap chorde's futures with it.
    """
    class WrapperFuture(base):
        def __init__(self, wrapped):
            self.__wrapped = wrapped

        for name, fn in vars(base).items():
            if not name.startswith('__') and callable(fn):
                def mkf(name, fn):
                    @functools.wraps(fn)
                    def f(self, *p, **kw):
                        return getattr(self.__wrapped, name)(*p, **kw)
                    return f
                locals()[name] = mkf(name, fn)
        del name, fn
    return WrapperFuture

class AsyncCacheProcessorThreadPool(ThreadPool):
    class AsyncCacheProcessorThread(ThreadPool.Process):
        pass
    Process = AsyncCacheProcessorThread

    def __init__(self, workers):
        if ThreadPool is multiprocessing.pool.ThreadPool:
            # This patches ThreadPool, which is broken when instanced
            # from inside a DummyThread (happens after forking)
            current = multiprocessing.dummy.current_process()
            if not hasattr(current, '_children'):
                current._children = weakref.WeakKeyDictionary()
        ThreadPool.__init__(self, workers)

class ProcessorStats(object):
    __slots__ = (
        'tasks_queued',
        'tasks_started',
        'tasks_cancelled',
        'tasks_completed',
        'wait_time_sum',
        'wait_time_min',
        'wait_time_max',
        'task_time_sum',
        'task_time_min',
        'task_time_max',
    )
    def __init__(self):
        self.reset()

    def reset(self):
        self.tasks_queued = self.tasks_started = self.tasks_cancelled = self.tasks_completed = 0
        self.wait_time_sum = self.task_time_sum = 0
        self.wait_time_min = self.wait_time_max = self.task_time_min = self.task_time_max = None

    def on_task_queued(self):
        self.tasks_queued += 1

    def on_task_cancelled(self):
        self.tasks_cancelled += 1

    def on_task_started(self, wait_time):
        self.tasks_started += 1
        self.wait_time_sum += wait_time
        wait_time_min = self.wait_time_min
        if wait_time_min is None or wait_time_min > wait_time:
            self.wait_time_min = wait_time
        wait_time_max = self.wait_time_max
        if wait_time_max is None or wait_time_max < wait_time:
            self.wait_time_max = wait_time

    def on_task_completed(self, task_time):
        self.tasks_completed += 1
        self.task_time_sum += task_time
        task_time_min = self.task_time_min
        if task_time_min is None or task_time_min > task_time:
            self.task_time_min = task_time
        task_time_max = self.task_time_max
        if task_time_max is None or task_time_max < task_time:
            self.task_time_max = task_time

class AsyncCacheProcessor(object):
    """
    An async cache processor will allow asynchronous reads
    and writes to a cache, Efficiently fitting into an async
    framework by passing and invoking callbacks.

    It modifies the cache interface to return a Future
    instead of a value, upon which an on_value(callback)
    method will retrieve the result, if any.

    If there is a cache miss, on_miss callbacks will be invoked
    instead, and in case of an exception, on_exc.

    Futures also implement Python 3's concurrent.futures.Future
    interface, see Future's documentation for more details.

    It also provides a do_async, that lets you dump arbitrary
    tasks on this processor's async processing pool (in case
    you need to do it for synchronization)
    """

    def __init__(self, workers, client,
            coalescence_buffer_size = 500, maxqueue = None, cleanup_cycles = 500,
            threadpool = None):
        self.client = client
        self.logger = logging.getLogger("chorde")
        self.workers = workers
        self.maxqueue = maxqueue
        self._wself = weakref.ref(self)
        if callable(threadpool):
            self._threadpool_factory = threadpool
            self._threadpool = None
        else:
            self._threadpool_factory = None
            self._threadpool = threadpool
        self._spawnlock = threading.Lock()

        self.coalesce_get = Cache(coalescence_buffer_size)
        self.coalesce_getTtl = Cache(coalescence_buffer_size)
        self.coalesce_contains = Cache(coalescence_buffer_size)

        self.tl = threading.local()
        self.cleanup_tasks = []
        self.cleanup_cycles = cleanup_cycles

        self._tit_tat = itertools.cycle(iter((True,False))).__next__

        self.stats = ProcessorStats()

    @property
    def threadpool(self):
        if self._threadpool is None:
            with self._spawnlock:
                if self._threadpool is None:
                    if self._threadpool_factory is not None:
                        self._threadpool = self._threadpool_factory(self.workers)
                    else:
                        self._threadpool = AsyncCacheProcessorThreadPool(self.workers)
        return self._threadpool

    def _enqueue(self, action, coalesce = None, coalesce_key = NONE):
        logger = self.logger
        cfuture = future = Future(logger)

        do_coalescence = coalesce is not None and coalesce_key is not NONE
        if do_coalescence:
            waction = weakref.ref(action)
            cfuture, cwaction = coalesce.setdefault(coalesce_key, (future, waction))
            if cfuture is not future and cwaction() is None:
                # Dead action, discard, reinsert
                cfuture = future
                try:
                    del coalesce[coalesce_key]
                except:
                    pass
                cfuture, cwaction = coalesce.setdefault(coalesce_key, (future, waction))
                logger.warning("Found zombie task in coalesce buffer with key %r", coalesce_key)

        if cfuture is future:
            maxqueue = self.maxqueue
            if maxqueue is not None and self.queuelen > (maxqueue*2):
                # Stop filling it
                if do_coalescence:
                    try:
                        del coalesce[coalesce_key]
                    except:
                        pass
                cfuture.cancel()
            else:
                # I'm the one queueing
                wself = self._wself
                stats = self.stats
                queue_time = time.time()
                def wrapped_action():
                    def clean_coalesce():
                        if do_coalescence:
                            try:
                                del coalesce[coalesce_key]
                            except:
                                pass

                    def clean(self):
                        if self is not None:
                            try:
                                tl = self.tl
                                if not hasattr(tl, 'dirty_rounds'):
                                    tl.dirty_rounds = 0
                                tl.dirty_rounds += 1
                                if tl.dirty_rounds > self.cleanup_cycles or not self.queuelen:
                                    tl.dirty_rounds = 0
                                    for task in self.cleanup_tasks:
                                        task()
                                    for task in _global_cleanup_tasks:
                                        task()
                            except:
                                logger.error("Error during background thread cleanup", exc_info = True)

                    # discard queue head quickly when we're overloaded
                    # head is always less relevant
                    self = wself()
                    cleaned_coalesce = False
                    try:
                        if self is not None and maxqueue is not None:
                            if self.queuelen > maxqueue:
                                # Only discard half the entries, otherwise we can
                                # enter a race condition in which really fast input
                                # to the processor keeps the queue full and discarding all
                                if self._tit_tat():
                                    cfuture.cancel()

                        if cfuture.set_running_or_notify_cancelled():
                            start_time = time.time()
                            stats.on_task_started(start_time - queue_time)
                            try:
                                rv = action()
                                clean_coalesce()
                                cleaned_coalesce = True
                                cfuture.set(rv)
                            except CacheMissError:
                                clean_coalesce()
                                cleaned_coalesce = True
                                cfuture.miss()
                            except:
                                clean_coalesce()
                                cleaned_coalesce = True
                                # Clear up traceback to avoid leaks
                                cfuture.exc(sys.exc_info()[:-1] + (None,))
                            stats.on_task_completed(time.time() - start_time)
                        else:
                            clean_coalesce()
                            cleaned_coalesce = True
                            stats.on_task_cancelled()
                    finally:
                        # Just in case, we really need to clean, or we leak cfutures
                        if not cleaned_coalesce:
                            clean_coalesce()
                            cleaned_coalesce = True
                        clean(self)
                self.threadpool.apply_async(wrapped_action, ())
                stats.on_task_queued()
        return cfuture

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    @property
    def queuelen(self):
        return self._threadpool._taskqueue.qsize() if self._threadpool is not None else 0

    def do_async(self, func, *args, **kwargs):
        return self._enqueue(functools.partial(func, *args, **kwargs))

    def do_async_coalescent(self, coalesce, coalesce_key, func, *args, **kwargs):
        """
        Returns an already-queued future if there is one, or a new one if not.

        Params:
            coalesce: A coalescence map (dict, Cache, or simialr)
            coalesce_key: A coalesence key

            See do_async for the rest
        """
        return self._enqueue(functools.partial(func, *args, **kwargs),
            coalesce, coalesce_key)

    def bound(self, client):
        """
        Returns a proxy of this processor bound to the specified client instead.
        """
        return WrappedCacheProcessor(self, client)

    def getTtl(self, key, default = NONE, **kw):
        if not kw or not (kw.keys() - COALESCE_IGNORE_KWARGS):
            if default is NONE:
                ckey = key
            else:
                ckey = (key, default)
        else:
            ckey = NONE
        return self._enqueue(functools.partial(self.client.getTtl, key, default, **kw),
            self.coalesce_getTtl, ckey)

    def get(self, key, default = NONE, **kw):
        if not kw or not (kw.keys() - COALESCE_IGNORE_KWARGS):
            if default is NONE:
                ckey = key
            else:
                ckey = (key, default)
        else:
            ckey = NONE
        return self._enqueue(functools.partial(self.client.get, key, default, **kw),
            self.coalesce_get, ckey)

    def contains(self, key, *p, **kw):
        if not p and not kw:
            ckey = key
        elif not kw:
            ckey = (key, p)
        else:
            ckey = (key, p, frozenset(kw.items()))
        return self._enqueue(functools.partial(self.client.contains, key, *p, **kw),
            self.coalesce_contains, ckey)

    def put(self, key, value, ttl, **kw):
        return self._enqueue(functools.partial(self.client.put, key, value, ttl, **kw))

    def renew(self, key, ttl, **kw):
        return self._enqueue(functools.partial(self.client.renew, key, ttl, **kw))

    def add(self, key, value, ttl, **kw):
        return self._enqueue(functools.partial(self.client.add, key, value, ttl, **kw))

    def delete(self, key):
        return self._enqueue(functools.partial(self.client.delete, key))

    def expire(self, key):
        return self._enqueue(functools.partial(self.client.expire, key))

    def clear(self):
        return self._enqueue(self.client.clear)

    def purge(self):
        return self._enqueue(self.client.purge)

    def register_cleanup(self, task):
        """
        Register a callable that will take no arguments, and will
        be invoked every cleanup_cycles tasks in order to perform
        thread-local cleanup for this processor.
        """
        self.cleanup_tasks.append(task)

    @staticmethod
    def register_default_cleanup(task):
        """
        Register a callable that will take no arguments, and will
        be invoked every cleanup_cycles tasks in order to perform
        thread-local cleanup for all async processors (including
        async writers).
        """
        _global_cleanup_tasks.append(task)

class WrappedCacheProcessor(object):
    """
    Wraps an AsyncCacheProcessor, binding its interface to a
    different client.
    """
    def __init__(self, processor, client, coalescence_buffer_size = 500):
        self.processor = processor
        self.client = client
        self.coalesce_get = Cache(coalescence_buffer_size)
        self.coalesce_getTtl = Cache(coalescence_buffer_size)
        self.coalesce_contains = Cache(coalescence_buffer_size)

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    @property
    def queuelen(self):
        return self.processor.queuelen

    @property
    def maxqueue(self):
        return self.processor.maxqueue

    @property
    def stats(self):
        return self.processor.stats

    def do_async(self, func, *args, **kwargs):
        return self.processor.do_async(func, *args, **kwargs)

    def bound(self, client):
        if client is self.client:
            return self
        else:
            return WrappedCacheProcessor(self.processor, client)

    def getTtl(self, key, default = NONE, **kw):
        if not kw:
            if default is NONE:
                ckey = key
            else:
                ckey = (key, default)
        else:
            ckey = NONE
        return self.processor.do_async_coalescent(self.coalesce_getTtl, ckey,
            self.client.getTtl, key, default, **kw)

    def get(self, key, default = NONE):
        if default is NONE:
            ckey = key
        else:
            ckey = (key, default)
        return self.processor.do_async_coalescent(self.coalesce_get, ckey,
            self.client.get, key, default)

    def contains(self, key, *p, **kw):
        if not p and not kw:
            ckey = key
        elif not kw:
            ckey = (key, p)
        else:
            ckey = (key, p, frozenset(kw.items()))
        return self.processor.do_async_coalescent(self.coalesce_contains, ckey,
            self.client.contains, key, *p, **kw)

    def put(self, key, value, ttl, **kw):
        return self.processor.do_async(self.client.put, key, value, ttl, **kw)

    def renew(self, key, ttl, **kw):
        return self.processor.do_async(self.client.renew, key, ttl, **kw)

    def add(self, key, value, ttl, **kw):
        return self.processor.do_async(self.client.add, key, value, ttl, **kw)

    def delete(self, key):
        return self.processor.do_async(self.client.delete, key)

    def expire(self, key):
        return self.processor.do_async(self.client.expire, key)

    def clear(self):
        return self.processor.do_async(self.client.clear)

    def purge(self, *p, **kw):
        return self.processor.do_async(self.client.purge, *p, **kw)

    def register_cleanup(self, task):
        self.processor.register_cleanup(task)

    @staticmethod
    def register_default_cleanup(task):
        AsyncCacheProcessor.register_default_cleanup(task)
